#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a directory containing FHI-AIMS input and output files."""

import os

import numpy as np

from PDielec import Constants
from PDielec.GenericOutputReader import GenericOutputReader
from PDielec.UnitCell import UnitCell


class AimsOutputReader(GenericOutputReader):
    """Read the contents of a directory containing Aims input and output files.

    Inherits from :class:`~PDielec.GenericOutputReader.GenericOutputReader`

    Parameters
    ----------
    names : list
        A list of file names to be used.

    """

    def __init__(self, names):
        """Initialise a Aims output reader.

        Parameters
        ----------
        names : list
            A list of file names to be used.

        This routine reads files with fixed names:
                  geometry.in         contains details of the cells used
                  hessian.???.yaml    contains the dynamical matrix
                  BORN_PDIELEC        generated using aims-pdielec-born

        """        
        GenericOutputReader.__init__(self, names)
        self.type                    = "Aims output"
        self.units                   = { "length"          : "angstrom",
                                         "mass"            : "AMU",
                                         "force"           : None,
                                         "force_constants" : None, }
        return

    def _read_output_files(self):
        """Read the Phonopy files in the directory.

        reads the dynamical matrix from FHI-Aims
        reads the Born charges
        """
        # Calculate dynamical matrix
        energy_filename = self._outputfiles[0]
        masses_filename = self._outputfiles[1]
        geometry_filename = self._outputfiles[2]
        hessian_filename  = self._outputfiles[3]
        born_filename     = self._outputfiles[4]
        if os.path.exists(energy_filename):
            self.read_energy(energy_filename)
        if os.path.exists(masses_filename) and os.path.exists(geometry_filename):
            self.read_geometry(masses_filename,geometry_filename)
        if os.path.exists(hessian_filename):
            self.read_dynamical_matrix(hessian_filename)
        if os.path.exists(born_filename):
            self.read_born_file(self.nions,born_filename)
        return

    def read_energy(self, file):
        """Read and process the aims log file.

        Parameters
        ----------
        file : str
            The aims log file, usually aims.out

        Returns
        -------
        None

        Modifies
        --------
        self.final_energy_without_entropy float
        self.final_energies_without_entropy float

        """        
        # Read the aims.out file
        with open(file) as fd:
            for line in fd:
                if "  | Total energy of the DFT " in line:
                    self.final_energy_without_entropy = float(line.split()[-2])
                    self.final_energies_without_entropy.append(self.final_energy_without_entropy)
        return

    def read_geometry(self, masses_file, geometry_file):
        """Read and process the FHI-AIMS masses.*.dat and geometry.in file.

        Parameters
        ----------
        masses_file : str
            The filename containing the FHI-Aims masses, usually masses.*.dat
        geometry_file : str
            The filename containing the FHI-Aims geometry, usually geometry.in

        Returns
        -------
        None

        Modifies
        --------
        self.unit_cell as a list with a single unit cell in it

        """        
        # Read the masses
        masses = []
        with open(masses_file) as fd:
            for line in fd:
                masses.append(float(line.split()[0]))
        # Read the geometry
        with open(geometry_file) as fd:
            lines = fd.readlines()
        lattice = []
        coordinates = []
        symbols = []
        for line in lines:
            split_line = line.split()[1:]
            if "lattice_vector" in line:
                lattice.append([ float(f) for f in split_line[:3] ])
            elif "atom_frac" in line:
                cartesian = False
                coordinates.append([ float(f) for f in split_line[:3] ])
                symbols.append(split_line[3])
            elif "atom" in line:
                cartesian = True
                coordinates.append([ float(f) for f in split_line[:3] ])
                symbols.append(split_line[3])
        cell = UnitCell(lattice[0], lattice[1], lattice[2], units=self.units["length"])
        if cartesian:
            cell.set_xyz_coordinates(coordinates)
        else:
            cell.set_fractional_coordinates(coordinates)
        cell.set_element_names(symbols)
        cell.set_atomic_masses(masses)
        self.unit_cell = [ cell ]
        return

    def read_dynamical_matrix(self,filename):
        """Read and process the dynamical matrix from output files.

        This method reads the dynamical matrix from the specified
        output files calculates the Hessian matrix, converts its units, and computes the
        eigenvalues and eigenvectors to determine frequencies and mass-weighted normal modes.

        Parameters
        ----------
        filename : str
            The filename containing the hessian, usually of the form hessian.*.dat

        Returns
        -------
        None
            This method directly modifies the instance attributes `_old_masses`, `frequencies`, 
            and `mass_weighted_normal_modes` based on the data read and calculations performed.

        Raises
        ------
        FileNotFoundError
            If any of the specified output files cannot be found or opened.
        MemoryError
            If there is insufficient memory to compute the eigendecomposition.

        Notes
        -----
        - The dynamical matrix is processed to compute real-valued Hessian by considering only the real
          parts. It is then converted to the desired units before computing the eigenvalues and eigenvectors.
        - The method computes frequencies in THz by first converting the eigenvalues from the Hessian to
          the correct units, then calculates the square root of their absolute values, preserving the sign
          to distinguish between real and imaginary modes.
        - Mass-weighted normal modes are then computed for each ion based on the eigenvectors, and the 
          frequencies and normal modes are stored in corresponding instance attributes.

        """        
        hessian = np.loadtxt(filename)
        cell = self.unit_cell[-1]
        self._old_masses = cell.get_atomic_masses()
        self.nions = len(self._old_masses)
        self.species = cell.get_species()
        self.nspecies = len(self.species)
        elements = cell.get_element_names()
        # Calculate the number of ions / type
        # Determine the atom type list
        self.ions_per_type = [ 0 for _species in self.species ]
        for element in elements:
            species_index = self.species.index(element)
            self.ions_per_type[species_index] += 1
            self.atom_type_list.append(species_index)
        # Determine the mass of each type
        self.mass_per_type = []
        for species in self.species:
             element_index = elements.index(species)
             self.masses_per_type.append(self._old_masses[element_index])
        # Proceed
        self.unit_cells = [ cell ]
        self.ncells = 1
        self.volume = cell.getVolume("Angstrom")
        # Values of units in SI
        ev = 1.0/Constants.hartree2ev
        amu = Constants.amu
        angs = Constants.angs2bohr
        wavenumber = Constants.wavenumber
        #
        # Process the hessian
        # Mass weighting
        #
        mass_vector = np.zeros([0])
        for mass in self._old_masses:
            mass_vector = np.append(mass_vector, np.ones(3) * 1.0/np.sqrt(mass))
        mass_mat=mass_vector[:,np.newaxis]*mass_vector[np.newaxis,:]  #1/sqrt(mimj)
        hessian = hessian * mass_mat
        #
        # The FHI-Aims units for the hessian are ev/(amu*Ang*Ang)
        # PDielec works in atomic units 
        #
        conversion_factor = ev / (amu * angs * angs) 
        hessian *= conversion_factor
        # Symmetrise hessian
        hessian = (hessian.T + hessian)/2.0
        # Find its eigenvalues and eigen vectors
        eig_val, eig_vec = np.linalg.eigh(hessian)
        self.mass_weighted_normal_modes = []
        nmodes = 3*self.nions
        # Store the new frequencies, using the negative convention for imaginary modes
        frequencies_a = np.sqrt(np.abs(eig_val.real)) * np.sign(eig_val.real)
        self.frequencies = (frequencies_a/wavenumber).tolist()
        # Store the mass weighted normal modes
        for i in range(nmodes):
            mode = []
            n = 0
            for _j in range(self.nions):
                modea = [eig_vec[n][i], eig_vec[n+1][i], eig_vec[n+2][i]]
                n = n + 3
                mode.append(modea)
            self.mass_weighted_normal_modes.append(mode)
        # end for i
        return

    def read_born_file(self,natoms,filename):
        """Read the BORN_PDIELEC file in the current directory.

        The BORN_PDIELEC file was created by a phonopy helper routine such as phonopy-pdielec-born
        It contains the optical permittivity and the Born charges

        Parameters
        ----------
        natoms : int
            The number of atoms
        filename : str
            The filename (probably "BORN_PDIELEC")

        Modifies
        --------
        zerof_optical_dielectric : the zero frequency optical permittivity
        born_charges             : the born charges

        """        
        with open(filename) as fd:
            #
            # Skip a line
            #
            line = fd.readline()
            #
            # Read dielectric constant
            #
            line = fd.readline().split()
            if len(line) != 9:
                print("BORN file format of line 2 is incorrect")
                return
            self.zerof_optical_dielectric = np.reshape([float(x) for x in line], (3, 3))
            #
            # Read Born effective charge
            #
            self.born_charges = np.zeros((natoms, 3, 3), dtype="double")
            for i in range(natoms):
                line = fd.readline().split()
                if len(line) == 0:
                    print("Number of lines for Born effect charge is not enough.")
                    return
                if len(line) != 9:
                    print("BORN file format of line %d is incorrect" % (i + 3))
                    return
                self.born_charges[i] = np.reshape([float(x) for x in line], (3, 3))
        return



#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""SpreadSheetMamager module."""
import numpy as np
import xlsxwriter as xlsx


class SpreadSheetManager:
    """A manager for handling operations on an Excel workbook via xlsxwriter.

    This class provides methods for managing a workbook with various tabs
    (named worksheets), writing data to these worksheets, and handling complex
    data types specifically tailored for scientific or engineering applications.

    Parameters
    ----------
    filename : str
        The name of the file to which the workbook will be saved.

    Attributes
    ----------
    workbook : Workbook
        The workbook object from xlsxwriter this manager operates on.
    closed : bool
        Boolean flag indicating whether the workbook has been closed.
    tab_names : list of str
        A list containing the names of the tabs (worksheets) to be managed.
    worksheets : dict
        A dictionary mapping tab names to worksheet objects.
    positions : dict
        A dictionary mapping tab names to their current write position (row, column).
    max_col : dict
        A dictionary mapping tab names to the maximum column index used.
    max_row : dict
        A dictionary mapping tab names to the maximum row index used.
    opened : dict
        A dictionary indicating whether a tab (worksheet) has been opened for writing.
    name : str
        The name of the currently selected work tab.

    Methods
    -------
    openWorkSheet(tab)
        Opens a new worksheet for the given tab if not already opened.
    selectWorkSheet(name)
        Selects the specified worksheet for further operations.
    writeNextRow(items, row=None, col=None, check='')
        Writes a series of items to the next row of the currently selected worksheet.
    write(row, col, item)
        Writes an item to the specified location in the currently selected worksheet.
    delete()
        Deletes all content from the currently selected worksheet.
    close()
        Closes the workbook, finalizing it for output.

    """

    def __init__(self, filename):
        # Manage a spread sheet for PDielec / PDGui
        """Initialize the instance with the given filename.

        This constructor initializes an Excel workbook with predefined tab names and settings. It prepares the workbook for data entry and analysis across various predefined categories related to powder and crystal property analysis.

        Parameters
        ----------
        filename : str
            The name of the file to which the workbook will be saved.

        Attributes
        ----------
        workbook : xlsx.Workbook
            An instance of an Excel workbook.
        closed : bool
            A flag indicating whether the workbook is closed.
        tab_names : list
            A list of strings indicating the names of the workbook tabs.
        worksheets : dict
            A dictionary to keep track of worksheet objects by tab name.
        positions : dict
            A dictionary storing the positions of the last written row and column for each tab.
        max_col : dict
            A dictionary tracking the maximum column number used in each tab.
        max_row : dict
            A dictionary tracking the maximum row number used in each tab.
        opened : dict
            A dictionary indicating whether a tab is open for editing.
        name : str
            The name of the currently active worksheet.

        Notes
        -----
        This constructor also automatically opens the worksheet named 'Main' to ready it for data entry or manipulation.

        Examples
        --------
        >>> my_excel_file = ExcelFileHandler("example.xlsx")
        >>> my_excel_file.add_data_to_sheet("Main", data)

        """        
        self.workbook = xlsx.Workbook(filename)
        self.closed = False
        self.tab_names = ["Main", "Settings", "Scenarios",
                "Powder Molar Absorption (cells)",
                "Powder Molar Absorption (atoms)",
                "Powder Molar Absorption (mols)",
                "Powder Absorption",
                "Powder Real Permittivity",
                "Powder Imaginary Permittivity",
                "Powder ATR Reflectance",
                "Analysis",
                "Crystal R_p",
                "Crystal R_s",
                "Crystal T_p",
                "Crystal T_s",
                "Crystal A_p",
                "Crystal A_s",
                "Real Crystal Permittivity",
                "Imag Crystal Permittivity" ]
        self.worksheets = {}
        # Positions points to where we write to next
        self.positions  = {}
        self.max_col    = {}
        self.max_row    = {}
        self.opened     = {}
        for tab in self.tab_names:
            self.opened[tab] = False
        self.name = "Main"
        self.openWorkSheet(self.name)

    def openWorkSheet(self,tab):
        """Open a new worksheet in a workbook.

        Parameters
        ----------
        tab : str
            The name of the worksheet to be opened.

        Returns
        -------
        None

        Notes
        -----
        If the worksheet identified by `tab` is already opened, this method does nothing. 
        Otherwise, it creates a new worksheet with the name `tab`, initializes its position, 
        maximum column, maximum row, and marks it as opened within the workbook data structures 
        handled by the instance. 

        """        
        if self.opened[tab]:
            return
        self.worksheets[tab] = self.workbook.add_worksheet(tab)
        self.positions[tab] = (0,0)
        self.max_col[tab] = 0
        self.max_row[tab] = 0
        self.opened[tab] = True

    def selectWorkSheet(self,name):
        """Select or opens a worksheet by name.

        Parameters
        ----------
        name : str
            The name of the worksheet to select or open.

        Returns
        -------
        None

        Notes
        -----
        This method selects a worksheet if it is already opened. If the worksheet is not
        currently open, it attempts to open the worksheet by calling `openWorkSheet` with
        the worksheet name.

        """        
        self.name = name
        if not self.opened[name]:
            self.openWorkSheet(self.name)

    def writeNextRow(self,items, row=None, col=None, check=""):
        """Write a sequence of items as a row in a spread sheet, starting from a specified row and column into a grid structure.

        Parameters
        ----------
        items : list
            A list of items to be written. Items can be strings, numbers, or nested lists where each sub-item is written in subsequent columns.
        row : int, optional
            The row index to start writing from. If not specified, uses the current row position of the object.
        col : int, optional
            The column index to start writing from. If not specified, uses the current column position of the object.
        check : str, optional
            A check string written to the beginning of the row specified. Defaults to an empty string if not specified.

        Returns
        -------
        None

        Notes
        -----
        - If `col` is 0, a warning message is printed.
        - The `row` and `col` are updated as items are written, and `row` is incremented after writing all items.
        - For items that are lists, each element is written in subsequent columns. This is applied recursively for nested lists.
        - `self.positions` is a dictionary holding the current positions (row, column) for different names, and `self.name` accesses the current object's name.

        """        
        oldRow,oldCol = self.positions[self.name]
        if col is None:
            col = oldCol
        if row is None:
            row = oldRow
        if col == 0:
            print("We have a problem, col is 0")
        self.write(row,0,check)
        for item in items:
            if isinstance(item,list):
                for i in item:
                    if isinstance(i, list):
                        for j in i:
                            self.write(row,col,j)
                            col += 1
                    else:
                        self.write(row,col,i)
                        col += 1
            else:
                self.write(row, col, item)
                col += 1
        row += 1

    def write(self,row,col,item):
        """Write an item to a specific position in the active worksheet, handling complex numbers.

        Parameters
        ----------
        row : int
            The row index where the item will be written.
        col : int
            The column index where the item will be written.
        item : int, float, complex, str
            The item to be written to the worksheet. If the item is a complex number,
            its real and imaginary parts are written to adjacent columns.

        Returns
        -------
        None

        Notes
        -----
        - This method updates the worksheet's max column and row based on the provided indices.
        - When writing a complex number, `col` is incremented before writing the imaginary part,
          and the updated column is used for adjusting the max column.
        - Updates the `positions` attribute with the next available position in the worksheet.

        """        
        if isinstance(item,complex) or np.iscomplexobj(item):
            self.worksheets[self.name].write(row,col,item.real)
            col += 1
            self.worksheets[self.name].write(row,col,item.imag)
            self.max_col[self.name] = max(self.max_col[self.name], col)
            self.max_row[self.name] = max(self.max_row[self.name], row)
            self.positions[self.name] = (row+1,col+1)
        else:
            self.worksheets[self.name].write(row,col,item)
            self.max_col[self.name] = max(self.max_col[self.name], col)
            self.max_row[self.name] = max(self.max_row[self.name], row)
            self.positions[self.name] = (row+1,col+1)

    def delete(self):
        """Delete all the data in the current worksheet.

        Clears all entries in the worksheet represented by `self` by setting each cell's value to an empty string.
        After clearing the data, it resets the current position, maximum column, and maximum row counters for the worksheet.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """        
        for row in range(0,self.max_row[self.name]):
            for col in range(0, self.max_col[self.name]):
                self.worksheets[self.name].write(row,col,"")
        self.positions[self.name] = (-1,0)
        self.max_col[self.name] = 0
        self.max_row[self.name] = 0

    def close(self):
        """Close the workbook if it hasn't been closed already.

        Closes the workbook associated with an instance and sets its status as closed to prevent multiple close operations.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """        
        if not self.closed:
            self.workbook.close()
        self.closed = True



#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""VASP output reader."""

import re
import xml.etree.ElementTree as ET

import numpy as np

from PDielec.Calculator import calculate_normal_modes_and_frequencies
from PDielec.Constants import atomic_number_to_element, hertz
from PDielec.GenericOutputReader import GenericOutputReader
from PDielec.UnitCell import UnitCell


def read_xml_element(ele):
    """Read an xml element and return a value.

    Parameters
    ----------
    ele : an xml element
        An xml element

    Returns
    -------
    name : str
        The name of the values returned
    values : float, int or bool (or a list of them)
        The return type depends on the contents of the element

    """
    if ele.tag == "separator":
        return None, None
    if ele.tag == "parameters":
        return None, None
    name = ele.attrib["name"]
    values = []
    if "type" in ele.attrib and ele.attrib["type"] == "int":
        values = [ int(f) for f in ele.text.strip().split() ]
    elif "type" in ele.attrib and ele.attrib["type"] == "logical":
        values = [ f == "T" for f in ele.text.strip().split() ]
    elif "type" in ele.attrib and ele.attrib["type"] == "string":
        values = [ None ] if ele.text is None else [ ele.text.strip() ]
    else:
        values = [ float(f) for f in ele.text.strip().split() ]
    if ele.tag == "i":
        values = values[0]
    return name,values


def myfloat(string):
    """Replace float(), return a large number if it finds a * in the string.

    Parameters
    ----------
    string : str
        The string to be analysed as a float

    Returns
    -------
    float
        A standard float conversion of the input or a large number if '*' is found in the input.

    """
    if "*" in string:
        return 9999.999
    return float(string)

class VaspOutputReader(GenericOutputReader):
    """Read the contents of a directory containing VASP input and output files.

    Inherits from :class:`~PDielec.GenericOutputReader.GenericOutputReader`

    Parameters
    ----------
    names : list of str
        A list of file names to be used in processing the DFT output files

    Attributes
    ----------
    type : str
        A string attribute that specifies the type of output, set to 'Vasp output'.
    _pspots : dict
        A private dictionary for the pseudo potentials
    _ediff : float
        A private float for the change in energy on each iteration
    _pulay : float
        The Pulay pressure
    _ibrion : int
        The value of ibrion
    _potim : float
        The value of potim

    """

    def __init__(self, names):
        """Initialize a new instance of the class.

        Parameters
        ----------
        names : list of str
            A list of file names to be used in processing the DFT output files

        Attributes
        ----------
        type : str
            A string attribute that specifies the type of output, set to 'Vasp output'.
        _pspots : dict
            A private dictionary for the pseudo potentials
        _ediff : float
            A private float for the change in energy on each iteration
        _pulay : float
            The Pulay pressure
        _ibrion : int
            The value of ibrion
        _potim : float
            The value of potim

        """        
        GenericOutputReader.__init__(self, names)
        self.type                    = "Vasp output"
        self._pspots                  = {}
        self._ediff                   = 0.0
        self._pulay                  = None
        self._ibrion                  = 0
        self._potim                   = 0.0
        return

    def _read_output_files(self):
        """Read the vasp files in the directory.

        Define the search strings neeeded and the associated routines will read the required values
        Process the files accordingly

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.manage = {}   # Empty the dictionary matching phrases
        self.manage["ionspertype"]  = (re.compile("   ions per type ="), self._read_ionspertype)
        # self.manage['masses_skip']  = (re.compile('  Mass of Ions in am'), self._read_skip4)
        self.manage["pspots"]       = (re.compile(" POTCAR:"), self._read_pspot)
        self.manage["arrays"]       = (re.compile(" Dimension of arrays:"), self._read_array_dimensions)
        #self.manage['masses']       = (re.compile('   POMASS ='), self._read_masses)
        self.manage["spin"]         = (re.compile("   ISPIN  = "), self._read_spin)
        self.manage["encut"]        = (re.compile("   ENCUT  = "), self._read_encut)
        self.manage["ediff"]        = (re.compile("   EDIFF  = "), self._read_ediff)
        self.manage["ibrion"]       = (re.compile("   IBRION = "), self._read_ibrion)
        self.manage["potim"]        = (re.compile("   POTIM  = "), self._read_potim)
        self.manage["nelect"]       = (re.compile("   NELECT = "), self._read_nelect)
        self.manage["lattice"]      = (re.compile("  volume of cell :"), self._read_lattice_vectors)
        self.manage["fractional"]   = (re.compile(" position of ions in fractional coordinates"), self._read_fractional_coordinates)
        self.manage["forces"]       = (re.compile(" POSITION  *TOTAL-FORCE"), self._read_forces)
        self.manage["energy"]       = (re.compile("  FREE ENERGIE OF THE ION"), self._read_energy)
        self.manage["magnet"]       = (re.compile(" number of electron "), self._read_magnet)
        self.manage["pressure"]     = (re.compile("  external pressure ="), self._read_external_pressure)
        self.manage["skip1"]        = (re.compile(" old parameters found"), self._read_skip4)
        self.manage["staticDielectric"]  = (re.compile(" MACROSCOPIC STATIC DIELECTRIC TENSOR .including"), self._read_static_dielectric)
        self.manage["staticIonic"]  = (re.compile(" MACROSCOPIC STATIC DIELECTRIC TENSOR IONIC CONT"), self._read_ionic_dielectric)
        self.manage["bornCharges"]  = (re.compile(" BORN EFFECTIVE CHARGES"), self._read_born_charges)
        self.manage["eigenvectors"] = (re.compile(" Eigenvectors and eige"), self._read_eigenvectors)
        self.manage["eigenskip"]    = (re.compile(" Eigenvectors after division"), self._read_skip4)
        self.manage["elastic"]      = (re.compile(" TOTAL ELASTIC MODULI"), self._read_elastic_constants)
        self.manage["kpoint"]      = (re.compile("^Gamma"), self._read_kpoint_grid)
        self.manage["species"]      = (re.compile("^ *Atomic configuration"), self._read_species)
        self.manage["newmasses"]  = (re.compile("  Mass of Ions in am"), self._read_newmasses)
        for f in self._outputfiles:
            if f.lower().endswith(".xml"):
                self._read_xml(f)
            else:
                self._read_output_file(f)
        return

    def _read_forces(self, line):
        """Read force data from a file and update the iterations datasets with maximum and RMS forces.

        After execution, `self.iterations` will be updated with the new `max_force` and `rms_force` values for the current iteration.

        Parameters
        ----------
        line : str
            The current line of the file to start reading from.

        Returns
        -------
        None

        """        
        line = self.file_descriptor.readline()
        maxf = 0.0
        rmsf = 0.0
        for _i in range(self.nions):
            line = self.file_descriptor.readline()
            forces = [ float(f) for f in line.split()[3:6] ]
            for f in forces:
                rmsf += f*f
                maxf = max(maxf, abs(f))
                # end if
            # end for f
        #end for i
        if "max_force" not in self.iterations:
            self.iterations["max_force"] = []
        self.iterations["max_force"].append(maxf)
        if "rms_force" not in self.iterations:
            self.iterations["rms_force"] = []
        self.iterations["rms_force"].append(rmsf)
        return

    def _read_species(self, line):
        """Read species from a file and updates species information.

        This method reads the number of lines for a species, computes the total charge from these lines, and appends the corresponding element to the species list. It utilizes the atomic_number_to_element dictionary to map the computed total charge to an element symbol.

        Parameters
        ----------
        line : str
            The initial line to start processing from, typically provides the number of lines for a species.

        Returns
        -------
        None

        See Also
        --------
        atomic_number_to_element : A dictionary mapping total charges to element symbols.

        """        
        line = self.file_descriptor.readline()
        nlines = int(line.split()[0])
        line = self.file_descriptor.readline()
        zcharge = 0.0
        for _i in range(nlines):
            line = self.file_descriptor.readline()
            zcharge = zcharge + float(line.split()[4])
        self.species.append(atomic_number_to_element[int(zcharge+0.001)])
        self.nspecies = len(self.species)
        return

    def _read_kpoint_grid(self, line):
        """Read and parse the k-point grid dimensions from a file line.

        This method assumes that the current line from which `line` is read contains three integers that represent the dimensions of the k-point grid. These integers are read, converted to integers, and stored in the instance variable `kpoint_grid`.

        Parameters
        ----------
        line : str
            The current line of the file to read from. This parameter is not used, and the actual line reading happens from `self.file_descriptor`.

        Returns
        -------
        None

        Notes
        -----
        - This method modifies the state of the object by setting the `kpoint_grid` attribute.
        - The method reads the next line from `self.file_descriptor`, not from the `line` parameter.

        """        
        line = self.file_descriptor.readline()
        self.kpoint_grid = [int(f) for f in line.split()[0:3] ]
        return

    def _read_ionspertype(self, line):
        """Read and store the number of ions per species from a given line.

        This method processes a string that contains information about the number of ions for each type/species. It updates the attributes `ions_per_type` and `nspecies` of the class.

        Parameters
        ----------
        line : str
            The line from which the ions per type information will be extracted. The relevant data is expected to start from the fifth element when splitting the line by spaces.

        Returns
        -------
        None

        """        
        self.ions_per_type = [int(i) for i in line.split()[4:]]
        self.nspecies = len(self.ions_per_type)
        return

    def _read_newmasses(self, line):
        """Read and process the "newmasses" from a line in the file descriptor.

        This internal method parses a single line from the file descriptor to read the masses associated with different atom types in a molecular system. Based on the number of species (`nspecies`) and ions per type, it updates `masses_per_type`, `masses`, `atom_type_list`, and `species_list` lists with the parsed and computed information.

        Parameters
        ----------
        line : str
            Line is a string (not used)

        Returns
        -------
        None
            The string line from which the new masses information is to be read. Note that this parameter is initially passed but not used directly, as the function immediately reads a new line from `self.file_descriptor`.

        """        
        self.masses_per_type = []
        line = self.file_descriptor.readline()
        mass_string = line[12:]
        start = 0
        increment = 6
        for _i in range(self.nspecies):
            mass = mass_string[start:start+increment]
            self.masses_per_type.append(float(mass))
            start = start + increment
        self.masses = []
        self.atom_type_list = []
        self.species_list = []
        for k, mass in enumerate(self.masses_per_type):
            n = self.ions_per_type[k]
            for _i in range(n):
                self.atom_type_list.append(k)
                self.masses.append(mass)
                self.species_list.append(self.species[k])
            # end loop over i
        # end look over current know types
        return

    def _read_masses(self, line):
        """Parse and store the mass value from a line of text.

        This method extracts the third element from a whitespace-separated list (assumed
        to be a mass), removes semicolons, converts it to a float, and appends it to the 
        instance's list of masses.

        Parameters
        ----------
        line : str
            A string from which the mass value is to be extracted. It is expected to
            be whitespace-separated with the mass as the third element, potentially
            followed by a semicolon.

        Returns
        -------
        None

        """        
        mass_string = line.split()[2]
        mass_string = mass_string.replace(";", "")
        self.masses_per_type.append(float(mass_string))
        return

    def _read_eigenvectors(self, line):
        """Read and process eigenvectors from a file.

        This method extracts frequencies and mass-weighted normal modes from a read file, storing them in the class's frequencies and mass_weighted_normal_modes attributes, respectively. It assumes the file follows a specific format where frequencies and their corresponding eigenvectors are listed, considering imaginary frequencies if present.

        Parameters
        ----------
        line : str
            This method operates on the instance's file_descriptor and nions attributes, assuming they are correctly initialized.

        Returns
        -------
        None

        """        
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.frequencies = []
        self.mass_weighted_normal_modes = []
        n = 3 * self.nions
        for _i in range(n):
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            imaginary = (line.split()[1] == "f/i=")
            freq = -float(line.split()[6]) if imaginary else float(line.split()[7])
            # end if
            self.frequencies.append(freq)
            line = self.file_descriptor.readline()
            a = []
            for _j in range(self.nions):
                line = self.file_descriptor.readline()
                a.append([float(line.split()[3]), float(line.split()[4]), float(line.split()[5])])
            # end for j
            self.mass_weighted_normal_modes.append(a)
        # end of for i in range(n)
        return

    def _read_born_charges(self, line):
        """Read the born charges from the OUTCAR file.

        Each row of the output refers to a given field direction and each column in the row refers to the atomic displacement. The output is arranged as: ::

            [[a1x a1y a1z]
             [a2x a2y a2z]
             [a3x a3y a3z]]

        where 1, 2, 3 are the field directions and x, y, z are the atomic displacements.

        Parameters
        ----------
        line : str 
            Not used and overridden

        Returns
        -------
        numpy.ndarray
            A numpy array containing the born charges with its shape determined by the field directions and atomic displacements as described above.

        """
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.born_charges = []
        for _i in range(self.nions):
            line = self.file_descriptor.readline()
            b = []
            b.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])
            line = self.file_descriptor.readline()
            b.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])
            line = self.file_descriptor.readline()
            b.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])
            line = self.file_descriptor.readline()
            self.born_charges.append(b)
        return

    def _read_elastic_constants(self, line):
        """Read and process elastic constants from a file.

        This method reads lines from an opened file pointed by the instance's file
        descriptor, extracts numeric values assumed to be elastic constants, and
        stores a processed version of these constants in the instance attribute
        `elastic_constants`.

        Parameters
        ----------
        line : str
            The initial line from where to start reading the elastic constants in
            the file. This parameter is somewhat misleading in this context, as the
            function does not actually use this input variable but rather immediately
            overrides it with `self.file_descriptor.readline()`.

        Returns
        -------
        None

        Notes
        -----
        Each line of interest contains numeric values, starting from the second
        value up to the seventh one, inclusive. These values are converted to floats,
        scaled down by a factor of 10 (to convert to GPa), assembled into a 2D array (list of lists),
        converted into a NumPy array, scaled, and then stored as a list in the
        `self.elastic_constants` attribute.

        """        
        # Read the total elastic constants
        elastic_constants = []
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        econs = np.array(elastic_constants)
        # convert from kBar to GPa
        econs = econs / 10.0
        self.elastic_constants = econs.tolist()
        return

    def _read_ionic_dielectric(self, line):
        """Read ionic dielectric constants from a file and update the zero-frequency static dielectric constant.

        This method assumes the presence of two successive lines in the file corresponding to two dielectric constant values. If any of these lines are marked with an asterisk (*) as missing or do not contain at least three numeric values, default values are used instead. The method calculates the sum of the existing zero-frequency optical dielectric constants and the newly read ionic dielectric constants and updates the corresponding class attribute with this sum.

        Parameters
        ----------
        line : str
            The initial line from which to start reading the ionic dielectric constants. This parameter is not directly used as the method starts by reading a new line from the file, but it's implicitly assumed to represent the current position in the file.

        Returns
        -------
        None

        Notes
        -----
        - Requires that `self.file_descriptor` is an open file object positioned at the correct line where dielectric information starts.
        - The method directly modifies `self.zerof_static_dielectric` based on read values and calculated sums.
        - This method performs no direct output but modifies the class state.

        """        
        # Read the ionic contribution to the static dielectric and use it to computet
        # the full static dielectric constant
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        # the is zero frequency ionic contribution to the static dielectric
        ionic_dielectric = []
        if "*" in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        ionic_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if "*" in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        ionic_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if "*" in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        ionic_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        array1 = np.array(self.zerof_optical_dielectric)
        array2 = np.array(ionic_dielectric)
        array3 = array1 + array2
        self.zerof_static_dielectric = array3.tolist()
        return

    def _read_static_dielectric(self, line):
        """Read and parse static dielectric data from a file.

        This method is designed to read lines from a file, which the instance's file_descriptor points to,
        and parse the static dielectric data. It expects the data to be in specific line positions within the file.
        If a line does not contain data in the expected format or contains an asterisk ('*') indicating missing
        or invalid data, a default value is used instead. The parsed data is stored in the instance's
        zerof_optical_dielectric attribute as a list of lists, where each sublist contains three floating-point numbers.

        Parameters
        ----------
        line : str
            The initial line from where to begin parsing. (This parameter is currently not used in the implementation
            and the method reads lines directly using the object's file_descriptor.)

        Returns
        -------
        None

        """        
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        # the is epsilon infinity
        self.zerof_optical_dielectric = []
        if "*" in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        self.zerof_optical_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if "*" in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        self.zerof_optical_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if "*" in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        self.zerof_optical_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        return

    def _read_skip4(self, line):
        """Read and skip the next 4 lines from the current position.

        This method reads and discards the next four lines from the file
        associated with the `file_descriptor` attribute of the calling object.

        Parameters
        ----------
        line : not used
            This parameter is not used but kept for function signature purposes.

        Returns
        -------
        None

        """        
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        return

    def _read_external_pressure(self, line):
        """Read and parse external pressure from a given line of text, updating instance attributes.

        Parameters
        ----------
        line : str
            A line of text expected to contain pressure data at specific positions.

        Returns
        -------
        None

        Notes
        -----
        This method updates the instance's `pressure` and `_pulay` attributes based on values extracted from the input string. The `pressure` value is obtained from the fourth item in the split string (`line.split()[3]`), and the `_pulay` value is from the ninth item (`line.split()[8]`). Both values are divided by 10.0 to convert from kbar to GPa.

        """        
        # Vasp writes out kbar so convert to GPa
        self.pressure = float(line.split()[3])/10.0
        self.pressures.append(float(line.split()[3])/10.0)
        self._pulay = float(line.split()[8])/10.0
        return

    def _read_pspot(self, line):
        """Read and store pseudopotential information from a line.

        This method parses a line to extract pseudopotential information and stores it in the `_pspots` dictionary attribute of the object. The pseudopotential label is used as the key, and its corresponding value as the value in the `_pspots` dictionary.

        Parameters
        ----------
        line : str
            A string containing the pseudopotential information to be parsed and stored. Expected format: "any_text value label any_other_text", where "value" and "label" are the parts of interest for storage.

        Returns
        -------
        None

        """        
        self._pspots[line.split()[2]] = line.split()[1]
        return

    def _read_array_dimensions(self, line):
        """Read and set the array dimensions from a line in a file.

        This method reads lines from a file pointed to by `self.file_descriptor`,
        extracts dimensions for kpoints, nbands, and nions from these lines, and
        sets the corresponding attributes of the class instance.

        Parameters
        ----------
        line : str
            The initial line read from the file, not used after modification.

        Returns
        -------
        None

        Notes
        -----
        Assumes specific formatting of the lines being read, where the required
        numbers are located at fixed positions in the lines:
        - `kpoints` is the 4th value in the first line read within the function.
        - `nbands` is the 15th value in the same line.
        - `nions` is the 12th value in the next line read.

        """        
        line = self.file_descriptor.readline()
        self.kpoints = int(line.split()[3])
        self.nbands = int(line.split()[14])
        line = self.file_descriptor.readline()
        self.nions = int(line.split()[11])
        return

    def _read_lattice_vectors(self, line):
        """Read and process lattice vectors from a line of a file.

        This method reads the volume and lattice vectors from specified lines of an input file,
        constructs a new `UnitCell` object from the read vectors, and appends it to the unit cells list.
        If there is at least one `UnitCell` already present, it copies fractional coordinates and element
        names from the last `UnitCell` in the list to the new one.

        Parameters
        ----------
        line : str
            The line from the input file containing the volume of the UnitCell.

        Returns
        -------
        None

        Attributes Modified
        -------------------
        volume : float
            The volume of the UnitCell, updated by parsing the input line.
        volumes : list of float
            List of volumes, appended with the current UnitCell's volume.
        unit_cells : list of UnitCell
            List of `UnitCell` objects, appended with the newly constructed `UnitCell`.

        """        
        self.volume = float(line.split()[4])
        self.volumes.append(float(line.split()[4]))
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        avector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        bvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        cvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        cell = UnitCell(avector, bvector, cvector,units="Angstrom")
        if self.ncells > 0:
            cell.set_fractional_coordinates(self.unit_cells[-1].fractional_coordinates)
            cell.set_element_names(self.species_list)
        self.unit_cells.append(cell)
        self.ncells = len(self.unit_cells)
        return

    def _read_fractional_coordinates(self, line):
        """Read fractional coordinates from a file and set them for the current unit cell.

        This function reads `nions` lines from the `file_descriptor` attribute, extracts
        the first three space-separated values as fractional coordinates, and appends them
        to the `ions` list. It then sets the fractional coordinates and element names for
        the last unit cell in the `unit_cells` list.

        Parameters
        ----------
        line : str
            The line from which to start reading the fractional coordinates. This parameter
            is passed but not used directly in the function; instead, subsequent lines are
            read from the file.

        Returns
        -------
        None

        """        
        ions = []
        for _n in range(self.nions):
            line = self.file_descriptor.readline()
            ions.append([float(f) for f in line.split()[0:3]])
        self.unit_cells[-1].set_fractional_coordinates(ions)
        self.unit_cells[-1].set_element_names(self.species_list)
        return

    def _read_spin(self, line):
        """Parse and set the spin from a given line.

        Parameters
        ----------
        line : str
            The line of text from which the spin value will be extracted.

        Returns
        -------
        None

        Notes
        -----
        This method expects the spin value to be at the third position in the line. The method doesn't return any value but sets the object's `spin` attribute based on the extracted integer value.

        """        
        self.spin = int(line.split()[2])
        return

    def _read_encut(self, line):
        """Parse and set the energy cutoff value from a given line.

        Parameters
        ----------
        line : str
            A string containing the energy cutoff value. The energy cutoff value is the third element in the string when split by whitespace.

        Returns
        -------
        None

        Notes
        -----
        This method does not return any value. It sets the value of `energy_cutoff` attribute of the instance based on the input line.

        """        
        self.energy_cutoff = float(line.split()[2])
        return

    def _read_ediff(self, line):
        """Read and set the energy difference value from a given line string.

        Parameters
        ----------
        line : str
            A string containing the energy difference information, where the
            energy difference value is expected to be at the third position
            (index 2) when split by whitespace.

        Returns
        -------
        None

        """        
        self._ediff = float(line.split()[2])
        return

    def _read_ibrion(self, line):
        """Read and set the IBRION parameter from a given line.

        This method extracts the IBRION value from a formatted line and sets it to the object's _ibrion attribute.

        Parameters
        ----------
        line : str
            String to be processed

        Returns
        -------
        None
            The line of text containing the IBRION value, expected to be at the third position (index 2) when split by whitespace.

        """        
        self._ibrion = int(line.split()[2])
        return

    def _read_potim(self, line):
        """Parse and set the POTIM parameter from a line of text.

        This method extracts the POTIM parameter value from a provided string
        and sets it as a floating-point number within the object.

        Parameters
        ----------
        line : str
            A string typically representing a line from a file containing
            the POTIM parameter. The expected format is: key-value pairs
            separated by spaces with the POTIM value being the third element
            in the sequence (index 2 when zero-indexed).

        Returns
        -------
        None

        """        
        self._potim = float(line.split()[2])
        return

    def _read_nelect(self, line):
        """Read and set the number of electrons from a given line.

        Parameters
        ----------
        line : str
            A string representation possibly containing information about
            the number of electrons.

        Notes
        -----
        This method updates the object's 'electrons' attribute with the number of electrons
        extracted from the given line. It assumes that the number of electrons is indicated
        at the third position (index 2) in the list produced by splitting `line` on whitespace.
        It also assumes that this number can be represented as a float, but converts it to
        an integer before setting the 'electrons' attribute.

        Returns
        -------
        None

        """        
        self.electrons = int(float(line.split()[2]))
        return

    def _read_magnet(self, line):
        """Read magnetization properties from a line and update the object attributes.

        Parameters
        ----------
        line : str
            A string expected to contain information about electrons and,
            optionally, magnetization. The expected format is whitespace-separated
            values where the fourth value is the number of electrons and the sixth

        Returns
        -------
        None
            value, if present, represents the magnetization.

        Notes
        -----
        Modifies the object's `electrons` attribute by setting it to the value parsed
        from the line. If magnetization information is present in the line, it also
        updates the `magnetization` attribute with that value; otherwise, it sets
        `magnetization` to 0.0.

        """        
        self.electrons = float(line.split()[3])
        if len(line.split()) > 5:
            self.magnetization = float(line.split()[5])
        else:
            self.magnetization = 0.0

    def _read_energy(self, line):
        """Read the energy-related values from a file and update the instance attributes with these values.

        Parameters
        ----------
        line : str
            The current line in the file from which to start reading. This parameter is not used in the function as the function reads new lines itself.

        Returns
        -------
        None

        Notes
        -----
        This method directly modifies the instance attributes `final_free_energy`, `final_free_energies`, `final_energy_without_entropy`, and `final_energies_without_entropy` by reading values from the file associated with `file_descriptor`. The relevant energy values can be extracted from the 5th and 4th positions (zero-indexed) of the line text split by spaces, on specific lines read in sequence.

        """        
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.final_free_energy = float(line.split()[4])
        self.final_free_energies.append(float(line.split()[4]))
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.final_energy_without_entropy = float(line.split()[3])
        self.final_energies_without_entropy.append(float(line.split()[3]))
        return

    def _read_xml(self, filename):
        """Process the VASP xml files.

        Parameters
        ----------
        filename : str
            The file name of the xml file to be processed

        Notes
        -----
        The routine only handles the Vasp output file vasprun.xml

        """
        # For each file generate an xml tree to process.
        tree = ET.parse(filename)
        root = tree.getroot()
        if self.debug:
            print(f"_read_xml filename = {filename}",flush=True)
            print(f"_read_xml tree = {tree}",flush=True)
            print(f"_read_xml root = {root}",flush=True)
        self._handle_kpoints_xml(root.findall("kpoints"))
        self._handle_parameters_xml(root.findall("parameters"))
        self._handle_atominfo_xml(root.findall("atominfo"))
        self._handle_structures_xml(root.findall("structure"))
        self._handle_calculations_xml(root.findall("calculation"))

    def _handle_calculations_xml(self,calculations_xml):
        """Process the calculations xml child.

        Parameters
        ----------
        calculations_xml : a list of xml element
            A list of calculations_xml elements

        Set
        ---
        self.nions : int
            The number of atoms
        self.nspecies : int
            The number of atoms
        self.atom_type_list : list of ints
            A list of the atom types in the order they will be given in the coordinate list
        self.zerof_static_dielectric : 3x3 list of floats
            The ionic contribution to the static dielectric 
        self.zerof_optical_dielectric : 3x3 list of floats
            The zero frequency optical dielectric dielectric 
        self.final_free_energies : a list of floats
            A list of the free energies at each geometry optimisation
        self.final_energies_without_entropy : a list of floats
            A list of the energies without entropy at each geometry optimisation
        self.borncharges : a list of floats nionsx3x3
            A list of the energies without entropy at each geometry optimisation
        self.mass_weighted_normal_modes : ndarray floats (3*nions x n*nions)
            The mass weighted normal modes
        self.frequencies : ndarray floats (3*nions)
            The frequencies in cm-1

        """
        dipole = None
        hessian = None
        dielectric_tensor = None
        forces_list = []
        stress_list = []
        unit_cell_list = []
        dipole_list = []
        # Loop around the calculations
        for calcxml in calculations_xml:
            lastscf = calcxml.findall("scstep/energy")[-1]
            dipoles = calcxml.findall("scstep/dipoles")
            lastdipole = dipoles[-1] if dipoles else None
            # Energies from the last step in this calculation
            # Correction because Vasp does not include the DFT energy
            # properly in the calculate/energy entry
            de = (float(lastscf.find('i[@name="e_0_energy"]').text) -
                  float(lastscf.find('i[@name="e_fr_energy"]').text))
            free_energy = float(calcxml.find('energy/i[@name="e_fr_energy"]').text)
            energy = free_energy + de
            self.final_free_energies.append(energy)
            energy = float(calcxml.find('energy/i[@name="e_wo_entrp"]').text)
            self.final_energies_without_entropy.append(energy)
            # Unit cell
            unit_cell = self._get_unit_cell_from_xml(calcxml.find("structure"))
            unit_cell_list.append(unit_cell)
            # Forces
            forces = None
            xml = calcxml.find('varray[@name="forces"]')
            if xml is not None:
                forces = []
                for vxml in xml:
                    forces.append( [float(f) for f in vxml.text.split()] )
            forces_list.append(forces)
            # stresses 
            stress = None
            xml = calcxml.find('varray[@name="stress"]')
            if xml is not None:
                stress = []
                for vxml in xml:
                    stress.append( [float(f)*0.1 for f in vxml.text.split()])
            stress_list.append(stress)
            # dipoles
            if lastdipole is not None:
                xml = lastdipole.find('v[@name="dipole"]')
                if xml is not None:
                    dipole = [ float(f) for f in xml.text.split() ] 
            xml = calcxml.find('dipole/v[@name="dipole"]')
            if xml is not None:
                dipole = [float(f) for f in xml.text.split()]
            dipole_list.append(dipole)
            # DFPT properties (There should only be one of these in a collection of calculations)
            # epsilon ion
            xml = calcxml.find('varray[@name="epsilon_ion"]')
            if xml is not None:
                zerof_static_dielectric = []
                for vxml in xml:
                    zerof_static_dielectric.append( [ float(f) for f in vxml.text.split()] )
                # print('zerof_static_dielectric',zerof_static_dielectric,flush=True)

            # epsilon optical
            xml = calcxml.find('varray[@name="epsilon"]')
            if xml is not None:
                self.zerof_optical_dielectric = []
                for vxml in xml:
                    self.zerof_optical_dielectric.append( [ float(f) for f in vxml.text.split()] )
                # print('self.zerof_optical_dielectric',self.zerof_optical_dielectric,flush=True)

            # hessian tensor
            xml = calcxml.find('dynmat/i[@name="unit"]')
            unit = xml if xml is not None else None
            xml = calcxml.find('dynmat/varray[@name="hessian"]')
            if xml is not None:
                hessian = []
                for v in xml.iter():
                    if v.tag == "v":
                        hessian.append( [ float(f) for f in v.text.split() ] )
            # dielectric tensor (epsilon_scf)
            xml = calcxml.find('varray[@name="epsilon_scf"]')
            if xml is not None:
                dielectric_tensor = []
                for vxml in xml:
                    dielectric_tensor.append( [ float(f) for f in vxml.text.split()] )
            # dielectric tensor (dielectric_dft)
            xml = calcxml.find('varray[@name="dielectric_dft"]')
            if xml is not None:
                dielectric_tensor = []
                for vxml in xml:
                    dielectric_tensor.append( [ float(f) for f in vxml.text.split()] )
                # print('dielectric_tensor',dielectric_tensor,flush=True)
            xml = calcxml.find('array[@name="born_charges"]')
            if xml is not None:
                # Born effective charges
                #        The order used with PDielec is:
                #        d2E/dFxdX d2E/dFxdY d2E/dFxdZ        0 3 6
                #        d2E/dFydX d2E/dFydY d2E/dFydZ        1 4 7
                #        d2E/dFzdX d2E/dFzdY d2E/dFzdZ        2 5 8
                self.born_charges = []
                for element_xml in xml[1:]:  # 1. element = dimension
                    charges = []
                    for vxml in element_xml:
                        charges.append( [ float(f) for f in vxml.text.split() ] )
                    self.born_charges.append(charges)
               #  print('born_charges',self.born_charges,flush=True)
        #
        # Convert Hessian units, first to hertz, then to atomic units
        # NB. this won't work with VASP 6
        #
        evtoj = 1.60217733E-19    # Taken from VASP5.4
        amtokg = 1.6605402E-27    # Taken from VASP5.4
        conversion = -1e+24 if unit is not None else -evtoj / amtokg / (4 * np.pi * np.pi) / 1e-10 ** 2
        conversion *= hertz**2
        if hessian is not None:
            hessian = conversion * np.array(hessian)
            self.mass_weighted_normal_modes, self.frequencies = calculate_normal_modes_and_frequencies(hessian)
        return

    def _handle_structures_xml(self,structures_xml):
        """Process the structures xml child.

        Parameters
        ----------
        structures_xml : a list of xml element
            A list of structurexml elements

        Set
        ---
        self.ncells : int
            The number of unit cells read so far
        self.volume : float
            The volume of the last unit cell read
        self.unit_cells : list of unit_cells
            A list of the unit cells

        """
        # Create a dictionary of structures
        structures = dict()
        for elem in structures_xml:
            name = elem.attrib.get("name")
            structures[name] = elem
        # Process the initial position
        self.unit_cells.append(self._get_unit_cell_from_xml(structures["initialpos"]))
        # Book keeping
        self.ncells = len(self.unit_cells)
        self.volume = self.unit_cells[-1].getVolume("Angstrom")
        if self.debug:
            print(f"_handle_structure_xml: volume={self.volume}",flush=True)
        return

    def _get_unit_cell_from_xml(self,structure_xml):
        """Read the xml and return a unit cell with the dimensions and positions.

        Parameters
        ----------
        structure_xml : a structure xml element
            A structure xml element

        Returns
        -------
        cell : a unit_cell

        """
        cell = []
        for v in structure_xml.find("crystal/varray[@name='basis']"):
            cell.append([ float(f) for f in v.text.split()])
        positions = []
        for v in structure_xml.find("varray[@name='positions']"):
            positions.append([ float(f) for f in v.text.split()])
        unit_cell = (UnitCell(cell[0], cell[1], cell[2], units="Angstrom"))
        unit_cell.set_fractional_coordinates(positions)
        unit_cell.set_element_names(self.species_list)
        return unit_cell

    def _handle_atominfo_xml(self,atominfos_xml):
        """Process the atminfos xml child.

        Parameters
        ----------
        atominfos_xml : a list of xml element
            A list of atminfo xml elements

        Set
        ---
        self.nions : int
            The number of atoms
        self.nspecies : int
            The number of atoms
        self.atom_type_list : list of ints
            A list of the atom types in the order they will be given in the coordinate list
        self.species : a list strings
            The list of unique species in the calculation. ie the list of atom types
        self.species_list : a list strings
            The list of species for all atoms
        self.ions_per_tpye : a list ints
            The list of atom types, could be used to lookup self.species
        self.masses_per_type : a list floats
            The list of masses for each atom type
        self.masses : a list floats
            The list of masses for each atom in the unit cell

        """
        self.nions = int( atominfos_xml[-1].find("atoms").text )
        self.nspecies = int( atominfos_xml[-1].find("types").text )
        self.atom_type_list = []
        self.ions_per_type = [ 0 for i in range(self.nspecies) ]
        self.species_list = []
        self.species = []
        for entry in atominfos_xml[-1].find("array/[@name='atoms']/set"):
            species = entry[0].text.strip()
            index = int(entry[1].text.strip())
            self.species_list.append(species)
            if species not in self.species:
                self.species.append(species)
            self.atom_type_list.append(index-1)
            self.ions_per_type[index-1] += 1
        self.masses_per_type = []
        for entry in atominfos_xml[-1].findall("array/[@name='atomtypes']/set/rc"):
            self.masses_per_type.append(float( entry.findall("c")[2].text.strip()))
        self.masses = []
        for atom_type in self.atom_type_list:
            self.masses.append(self.masses_per_type[atom_type])
        if self.debug:
            print("_handle_atomicinfo_xml: nspecies", self.nspecies,flush=True)
            print("_handle_atomicinfo_xml: nions", self.nions,flush=True)
            print("_handle_atomicinfo_xml: ions_per_type", self.ions_per_type,flush=True)
            print("_handle_atomicinfo_xml: atom_type_list", self.atom_type_list,flush=True)
            print("_handle_atomicinfo_xml: species", self.species,flush=True)
            print("_handle_atomicinfo_xml: species_list", self.species_list,flush=True)
            print("_handle_atomicinfo_xml: masses", self.masses,flush=True)
            print("_handle_atomicinfo_xml: masses_per_type", self.masses_per_type,flush=True)
        return

    def _handle_parameters_xml(self,parameters_xml):
        """Process the parameters xml child.

        Parameters
        ----------
        parameters_xml : a list of xml element
            A list of parameters xml elements

        Set
        ---
        self.electrons : int
            The number of electrons
        self.spin : int
            The spin of the system

        """
        self.electrons = 0
        parameters_xml=parameters_xml[-1]
        parameters = dict()
        for ele in parameters_xml.iter():
            name,value = read_xml_element(ele)
            if name is not None:
                parameters[name] = value
        self.electrons = int(parameters["NELECT"]+0.0001)
        self.spin = parameters["ISPIN"]
        # self.masses_per_type = parameters["POMASS"] 
        if self.debug:
            print("_handle_parameters_xml: nelect", self.electrons,flush=True)
            print("_handle_parameters_xml: spin", self.spin,flush=True)
        return

    def _handle_kpoints_xml(self,kpoints_xml):
        """Process the kpoints xml child.

        Parameters
        ----------
        kpoints_xml : a list of xml element
            A list of kpoint xml elements

        Set
        ---
        self.kpoint_grid : a list of 3 integers
            The Monkorst Pack grid
        self.kpoints : int
            The number of kpoints

        """
        kpoints_xml=kpoints_xml[-1]
        self.kpoint_grid = [1, 1, 1]
        xml = kpoints_xml.find("generation").find("v/[@name='divisions']")
        self.kpoint_grid = [ int(f) for f in xml.text.split() ]
        if self.debug:
            print("_handle_kpoints_xml: kpoint_grid", self.kpoint_grid ,flush=True)
        # Number of kpoints
        xml = kpoints_xml.find("varray/[@name='kpointlist']")
        vs = xml.findall("v")
        self.kpoints = len(vs)
        if self.debug:
            print("_handle_kpoints_xml: kpoints", self.kpoints,flush=True)
        return



#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""Generic reader of output files. An actual reader should inherit from this class."""
import math
import os
import sys

import numpy as np

from PDielec.Calculator import calculate_normal_modes_and_frequencies, cleanup_symbol
from PDielec.Constants import amu, avogadro_si, wavenumber
from PDielec.IO import pdielec_io
from PDielec.Plotter import print3x3, print_ints, print_reals, print_strings


class GenericOutputReader:
    """Generic reader of output files. Actual reader should inherit from this class.

    Examples of readers with this base class:

    - :class:`~PDielec.AbinitOutputReader`
    - :class:`~PDielec.CastepOutputReader`
    - :class:`~PDielec.CrystalOutputReader`
    - :class:`~PDielec.ExperimentOutputReader`
    - :class:`~PDielec.GulpOutputReader`
    - :class:`~PDielec.PhonopyOutputReader`
    - :class:`~PDielec.QEOutputReader`
    - :class:`~PDielec.VaspOutputReader`

    Parameters
    ----------
    filenames : list
        A list containing file names (strings) to be associated with the instance. These
        are converted into the absolute paths and stored.

    Attributes
    ----------
    _outputfiles : list
        Stores the initial filenames provided during instantiation.

    names : list
        The absolute paths of the files specified in `filenames`. Computed by resolving
        each filename to its absolute path.

    debug : bool
        Indicates if debug mode is on or off. Defaults to False.

    type : str
        The type of the object, initially set as 'Unkown'.

    ncells, nsteps, electrons, spin, nbands, volume, nions, nspecies, geomsteps : int
        Attributes initialized as 0, representing various numerical properties.

    unit_cells, volumes, species, energiesDFT, energiesDFT_disp, final_free_energies, final_energies_without_entropy, volumes, pressures, born_charges, frequencies, mass_weighted_normal_modes, ions_per_type, atom_type_list, masses, masses_per_type, elastic_constants, zerof_optical_dielectric, zerof_static_dielectric : list
        Various lists to store computational results related to the object. Initialized as empty lists.
    
    final_free_energy, final_energy_without_entropy, magnetization, energy_cutoff, pressure : float
        Floating point attributes initialized to represent different scalar quantities. Defaults to 0 or 0.0.

    kpoints : int
        The number of kpoints, initialized to 1.

    kpoint_grid : list
        The kpoint grid dimensions, initialized as [1, 1, 1].

    manage : set
        The set contains phrases which are matched when looking through a file

    iterations : dict
        Holds the values of 'max force' and 'rms force' at each iteration

    file_descriptor : str
        A string to hold file descriptor, initialized as an empty string.

    eckart : bool
        A boolean flag to indicate the use of Eckart conditions, defaults to False.

    hessian_symmetrisation : str
        The method for symmetrization of Hessian, defaults to 'symm'.

    open_filename, open_directory : str
        Strings for tracking open files or directories, initialized as empty strings.

    _old_masses : list
        A list to keep track of old mass values, initialized as empty.

    nomass_hessian : type or None
        Placeholder for hessian without mass contribution, initially None.

    nomass_hessian_has_been_set : bool
        Indicates if the `nomass_hessian` has been provided, defaults to False.

    original_born_charges, CrystalPermittivity, oscillator_strengths, edited_masses : Various types
        Attributes initialized as `None` to represent unassigned computational properties.

    original_born_charges_are_being_used : bool
        Signifies whether original Born charges are in use, defaults to True.

    """

    def __init__(self, filenames):
        """Initialise an instance of the reader.

        Parameters
        ----------
        filenames : list
            A list containing file names (strings) to be associated with the instance. These
            are converted into the absolute paths and stored.

        """        
        # Test to see if all the files needed are there
        # the test on "ignore" is present because of the FHI-Aims interface
        if not all( [ os.path.isfile(f) for f in filenames if "ignore" not in f ] ):
            print("Error in calling the generic output reader some files are not present")
            print(filenames)
            return
        self._outputfiles               = filenames
        self.names                      = [os.path.abspath(f) for f in filenames]
        self.debug                      = False
        self.type                       = "Unkown"
        self.ncells                     = 0
        self.unit_cells                 = []
        self.nsteps                     = 0
        self.electrons                  = 0
        self.spin                       = 0
        self.nbands                     = 0
        self.volume                     = 0
        self.volumes                    = []
        self.nions                      = 0
        self.nspecies                   = 0
        self.geomsteps                  = 0
        self.species                    = []       # a list of the unique species
        self.species_list               = []       # a list of the species for all atoms
        self.energiesDFT                = []
        self.energiesDFT_disp           = []
        self.final_free_energy          = 0.0
        self.final_free_energies        = []
        self.final_energy_without_entropy = 0.0
        self.final_energies_without_entropy = []
        self.kpoints                    = 1
        self.kpoint_grid                = [ 1, 1, 1 ]
        self.energy_cutoff              = 0.0
        self.born_charges               = []
        self.manage                     = {}
        self.iterations                 = {}
        self.file_descriptor            = ""
        self.pressure                   = 0
        self.pressures                  = []
        self.magnetization              = 0.0
        # this in epsilon infinity
        self.zerof_optical_dielectric   = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        # this is the zero frequency static dielectric constant
        self.zerof_static_dielectric    = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        self.elastic_constants          = [[0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0]]
        self.frequencies                = []
        self.mass_weighted_normal_modes = []
        self.ions_per_type              = []
        self.atom_type_list             = []
        self.masses                     = []
        self.masses_per_type            = []
        self.program_mass_dictionary    = {}
        self.eckart                     = False
        self.hessian_symmetrisation     = "symm"
        self.open_filename              = ""
        self.open_directory             = ""
        self._old_masses                = []
        self.nomass_hessian             = None
        self.nomass_hessian_has_been_set= False
        self.original_born_charges      = None
        self.original_born_charges_are_being_used = True
        self.CrystalPermittivity        = None
        self.oscillator_strengths       = None
        self.edited_masses              = None
        self.primitive_transformation   = None
        return

    def read_output(self):
        """Interface to the private read output files methods.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self._read_output_files()
        return

    def reset_masses(self):
        """Reset the mass values to the program's defaults.

        This function resets the mass dictionary of an object to the values specified
        by the object's `program_mass_dictionary`. Typically used when needing to revert
        any runtime changes to the object's mass values to their original state.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        If `debug` attribute of the object is True, it prints a message indicating 
        that the masses are being reset. This function does not return any value
        but modifies the object's state by changing its mass dictionary to match
        the `program_mass_dictionary`.

        Examples
        --------
        Assuming an object `obj` of a class that has `reset_masses` method, `debug` attribute,
        and `program_mass_dictionary`:

            >>> obj.debug = True
            >>> obj.reset_masses()
            Re setting mass dictionary to the program values
    
        Make sure that `program_mass_dictionary` is set properly before calling this
        method to avoid setting the masses to an unintended state.

        """        
        mass_dictionary = {}
        if self.debug:
            print("Re setting mass dictionary to the program values")
        if self.program_mass_dictionary:
            self.change_masses(self.program_mass_dictionary,mass_dictionary)

    def getSpecies(self):
        """Return a list of cleaned species symbols.

        This method applies a specified cleanup function to each element of 
        the `species` attribute and returns the resulting list.

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of cleaned species symbols.

        """        
        return [ cleanup_symbol(el) for el in self.species ]

    def mass_dictionary(self):
        """Generate a dictionary mapping chemical species symbols to their masses.

        Parameters
        ----------
        None

        Returns
        -------
        dictionary : dict
            A dictionary where keys are cleaned up species symbols and values are their corresponding masses.

        Notes
        -----
        The function uses a helper function `cleanup_symbol` to clean up the species symbols before using them as keys in the dictionary. If the instance's `debug` attribute is set to True, the function will print the newly created mass dictionary.

        Examples
        --------
        Assuming an object `molecule` with species `['H2', 'O']`, masses_per_type `[2.016, 15.999]`, and a debug attribute set to True, calling `molecule.mass_dictionary()` would print `{'H2': 2.016, 'O': 15.999}` and return this dictionary.

        """        
        dictionary = {}
        for symbol,mass in zip(self.species,self.masses_per_type):
            # the element name may be appended with a digit or an underscore
            element = cleanup_symbol(symbol)
            dictionary[element] = mass
        if self.debug:
            print("new mass_dictionary", dictionary)
        return dictionary

    def set_edited_masses(self,masses):
        """Set the edited masses attribute if the length matches the original masses attribute.

        Parameters
        ----------
        masses : list
            A list of mass values to be set as the edited masses.

        Returns
        -------
        None

        Notes
        -----
        The method updates the `edited_masses` attribute only if the length of the input list matches the length of the `masses` attribute of the object. If they do not match, an error message is printed, and `edited_masses` is set to `None`.

        Returns
        -------
        None

        """        
        if len(masses) == len(self.masses):
            self.edited_masses = masses
        else:
            print("Error unable to edited_masses")
            self.edited_masses = None
        return

    def change_masses(self, new_masses, mass_dictionary):
        """Change the masses of elements in a output reader.

        This function updates the masses of the elements within the reader according to the new masses provided. If the object has already had its masses edited, those edited masses will be used. Otherwise, it sets up or updates the simulation's internal mass dictionary based on the provided `new_masses` and an optional external `mass_dictionary` for overrides. If debugging is enabled, various diagnostic messages will be printed during the function's execution.

        Parameters
        ----------
        new_masses : dict
            A dictionary where keys are element symbols as strings and values are the new masses for these elements.
        mass_dictionary : dict, optional
            An additional dictionary provided for mass overrides. If any element present in `new_masses` also exists in `mass_dictionary`, the mass from `mass_dictionary` will be used.

        Returns
        -------
        None

        Notes
        -----
        - It assumes that `cleanup_symbol()` is a function used to standardize or validate the element symbols in `self.species`.
        - It alters `self.masses` and `self.masses_per_type` to reflect changes in masses as per `new_masses` and `mass_dictionary`.
        - If `self.edited_masses` is already set to a truthy value, it bypasses the update process and uses these values instead.
        - `self.species` and `self.atom_type_list` are expected to be iterable attributes of the object containing symbols for elements and types of atoms, respectively.
        - Debugging messages are conditionally printed based on the boolean attribute `self.debug`.

        """        
        if self.edited_masses:
            # This is pretty crude!  If the reader has this variable set then we
            # only use the masses stored in the edit_masses list
            # Once this variable has been set then only these masses are used
            if self.debug:
                print("Using the edited masses")
            self.masses = self.edited_masses
            return
        if not self.program_mass_dictionary:
            # We only want to do this once - remember the program masses as a dictionary
            if self.debug:
                print("Setting program mass dictionary")
            for symbol,mass in zip(self.species,self.masses_per_type):
                element = cleanup_symbol(symbol)
                self.program_mass_dictionary[element] = mass
        if self.debug:
            print("changing masses", self.program_mass_dictionary)
        self.masses = []
        self.masses_per_type = []
        for symbol in self.species:
            # the element name may be appended with a digit or an underscore
            element = cleanup_symbol(symbol)
            mass = new_masses[element]
            if element in mass_dictionary:
                mass = mass_dictionary[element]
            self.masses_per_type.append(mass)
        # end for symbol
        self.masses = [ self.masses_per_type[atype] for atype in self.atom_type_list ]
        if self.debug:
            print("new masses", self.masses)
        return

    def print(self):
        """Print information about the reader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        This function prints out detailed information about the reader it is run on.

        """
        # Generic printing of information
        print("")
        print("Summary of information contained in the QM/MM Reader")
        print("")
        print(f"Number of atoms: {self.nions:5d}")
        print("")
        print(f"Number of species: {self.nspecies:5d}")
        print_strings("Species:", self.species)
        print_ints("Number of atoms for each species:", self.ions_per_type)
        print_reals("Mass of each species:", self.masses_per_type,format="{:10.6f}")
        print_ints("Atom type list:", self.atom_type_list)
        print("")
        print(f"Number of kpoints: {self.kpoints:5d}")
        print("")
        print(f"Kpoint grid      : {self.kpoint_grid[0]:5d} {self.kpoint_grid[1]:5d} {self.kpoint_grid[2]:5d}")
        print("")
        print(f"Energy cutoff (eV): {self.energy_cutoff:f}")
        print("")
        print(f"final_free_energy(eV): {self.final_free_energy:f}")
        print("")
        print(f"geomsteps: {self.geomsteps:f}")
        print("")
        print_reals("DFT energies (eV):", self.energiesDFT,format="{:10.8f}")
        print("")
        print_reals("DFT energies including dispersion (eV):", self.energiesDFT_disp,format="{:10.8f}")
        print("")
        print_reals("Volumes:", self.volumes,format="{:10.8f}")
        print("")
        print_reals("Frequencies (cm-1):", self.frequencies)
        print_reals("Masses (amu):", self.masses,format="{:10.6f}")
        for i, charges in enumerate(self.born_charges):
            title = f"Born Charges for Atom {i:d}"
            print3x3(title, charges)
        print3x3("Epsilon inf: ", self.zerof_optical_dielectric)
        #jk print3x3("Unit cell: ", self.unit_cells[-1].lattice)
        self.unit_cells[-1].print()
        print(" ")
        print(f"Volume of cell: {self.volume:f}")
        mtotal = 0.0
        for m in self.masses:
            mtotal = mtotal + m
        print(f"Total mass is: {mtotal:f} g/mol")
        print(f"Density is: {mtotal/(avogadro_si * self.volume * 1.0e-24):f} g/cc")
        print(" ")
        return

    def get_unit_cell(self):
        """Return the last unit cell in the reader.

        The routine adds the current set of masses to the unit cell.
        It returns None if no unit cells have been read.

        Parameters
        ----------
        None

        Returns
        -------
        unitCell object
            The last unit cell read by the reader

        """
        if len(self.unit_cells) == 0:
            return None
        # Access the last unit cell in the reader
        cell = self.unit_cells[-1]
        # Add the current masses in the reader to the unit cell
        cell.set_atomic_masses(self.masses)
        # return the cell
        return cell

    def get_crystal_density(self):
        """Return the crystal density in g/cc.

        The volume is in angstrom^3, the masses are in atomic mass units, the density is in g/cc

        Parameters
        ----------
        None

        Returns
        -------
        float
            Density in g/cc

        """
        mtotal = 0.0
        for m in self.masses:
            mtotal = mtotal + m
        return mtotal/(avogadro_si * self.volume * 1.0e-24)

    def _read_output_files(self):
        """Read the through the output files.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        This method is over-ridden by the child class

        """
        # Define the search keys to be looked for in the files
        print("Error _read_output_files must be defined by the actual file reader")
        return

    def _read_output_file(self, name):
        """Read through the file 'name' for key words. The keywords are established in _read_output_files.

        Parameters
        ----------
        name : str
            The file name of the output file

        Returns
        -------
        None

        Notes
        -----
        The actual implementation for identifying keywords is handled by the `_read_output_files` method.

        """
        # Check to see if the file exists....
        if not os.path.isfile(name):
            print("Warning file is not present: ", name, file=sys.stderr)
            return
        # Open file and store file name and directory
        self.file_descriptor = pdielec_io(name, "r")
        self.open_filename = name
        self.open_directory = os.path.dirname(name)
        if self.open_directory == "":
            self.open_directory = "."
        # Loop through the contents of the file a line at a time and parse the contents
        line = self.file_descriptor.readline()
        while line != "":
            for k in self.manage:
                if self.manage[k][0].match(line):
                    method   = self.manage[k][1]
                    if self.debug:
                        print(f"_read_output_file({name}): Match found {k}")
                    method(line)
                    break
                # end if
            # end for
            line = self.file_descriptor.readline()
        # end while
        self.file_descriptor.close()
        return

    def _symmetric_orthogonalisation(self, A):
        """Private routine to perform symmetric orthogonalization.

        Parameters
        ----------
        A : ndarray
            matrix to be orthogonalised

        Returns
        -------
        None

        """
        # The matrix A is only approximately orthogonal
        n = np.size(A, 0)
        unity = np.eye(n)
        Ak = A
        for _k in range(3):
            Bk = np.dot(Ak, Ak.T)
            error = unity - Bk
            Ck = np.linalg.inv(unity + Bk)
            Kk = np.dot(error, Ck)
            Ak = np.dot((unity + Kk), Ak)
            error  = np.sum(np.abs(error))
            if self.debug:
                print("Orthogonalisation iteration: ", error)
        # end for k
        return Ak

    def calculate_mass_weighted_normal_modes(self):
        """Calculate the mass weighted normal modes from the hessian.
  
        The hessian itself is constructed from the frequencies and normal modes
        Any changes to the atomic masses is applied
        Eckart conditions are applied if requested

        Parameters
        ----------
        None

        Returns
        -------
        mass_weighted_normal_modes : np.array
           the mass weighted normal modes

        """
        #
        # Reconstruct the massweighted hessian
        # If necessary diagonalise it and extract the frequencies and normal modes
        # Any projection can be performed here
        #
        # First step is to reconstruct the dynamical matrix (D) from the frequencies and the eigenvectors
        # f^2 = UT . D . U
        # and U is a hermitian matrix so U-1 = UT
        # D = (UT)-1 f^2 U-1 = U f UT
        # Construct UT from the normal modes
        if self.debug:
            print("calculate mass weighted normal modes")
        n = np.size(self.mass_weighted_normal_modes, 0)
        m = np.size(self.mass_weighted_normal_modes, 1)*3
        UT = np.zeros((n, m))
        frequencies_a = np.array(self.frequencies) * wavenumber
        if self.debug:
            print("frequencies_a",frequencies_a)
        masses = np.array(self.masses)*amu
        # if the non mass-weighted hasn't been set, set it
        if not self.nomass_hessian_has_been_set:
            if self.debug:
                print("hessian was not set")
            self.nomass_hessian_has_been_set = True
            for imode, mode in enumerate(self.mass_weighted_normal_modes):
                n = 0
                for atom in mode:
                    # in python the first index is the row of the matrix, the second is the column
                    UT[imode, n+0] = atom[0]
                    UT[imode, n+1] = atom[1]
                    UT[imode, n+2] = atom[2]
                    n = n + 3
                # end for atom
            # end for imode
            # convert the frequencies^2 to a real diagonal array
            # Warning we have to make sure the sign is correct here
            # The convention is that if the frequency is negative
            # then it is really imaginary, so the square of the frequency
            # will be negative too.
            frequencies_a = np.array(self.frequencies) * wavenumber
            f2 = np.diag(np.sign(frequencies_a)*np.real(frequencies_a*frequencies_a))
            # The back transformation uses approximately orthogonal (unitary) matrices because of rounding issues on reading vectors
            # So before that lets orthogonalise them
            UT = self._symmetric_orthogonalisation(UT)
            hessian = np.dot(np.dot(UT.T, f2), UT)
            # Make sure the dynamical matrix is real
            hessian = np.real(hessian)
            # We are going to store the non mass-weighted hessian, since the frequencies came from the MM/QM program
            # they are calculated using the program masses
            current_mass_dictionary = self.mass_dictionary()
            if self.debug:
                print("current mass dictionary", current_mass_dictionary)
            # There is a chance that the program_mass_dictionary hasn't been set - if it hasn't use the current masses
            if not self.program_mass_dictionary:
                self.program_mass_dictionary = current_mass_dictionary
            if self.debug:
                print("program mass dictionary", self.program_mass_dictionary)
            self.change_masses(self.program_mass_dictionary, {})
            masses = np.array(self.masses)*amu
            # remove the mass weighting from the hessian and store
            self.nomass_hessian = self._remove_mass_weighting(hessian,masses)
            # finally replace the masses with those set before we did this
            self.change_masses(current_mass_dictionary, {})
            if self.debug:
                print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        # If the masses have been changed then alter the mass weighted hessian here
        masses = np.array(self.masses)*amu
        if self.debug:
            print("masses", masses)
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        self.hessian = self._modify_mass_weighting(self.nomass_hessian, masses)
        if self.debug:
            print("mass weighted hessian", self.hessian[0:4][0])
        # Project out the translational modes if requested
        if self.eckart:
            self.hessian = self.project(self.hessian)
            if self.debug:
                print("projected hessian", self.hessian[0:4][0])
        # Find its eigenvalues and eigen vectors
        self.mass_weighted_normal_modes, self.frequencies = calculate_normal_modes_and_frequencies(self.hessian)
        if self.debug:
            print("calculated frequencies", self.frequencies)
            print("mass-weighted normal modes", self.mass_weighted_normal_modes)
        # end for i
        return self.mass_weighted_normal_modes

    def project(self, hessian):
        """Apply projection operators to remove translation.

        Parameters
        ----------
        hessian : np.array
            The given matrix.

        Returns
        -------
        np.array
            The matrix with translational modes projected out.

        """
        new_hessian = np.zeros_like(hessian)
        nmodes = self.nions*3
        unit = np.eye(nmodes)
        p1 = np.zeros(nmodes)
        p2 = np.zeros(nmodes)
        p3 = np.zeros(nmodes)
        for i in range(self.nions):
            p1[i*3+0] = math.sqrt(self.masses[i])
            p2[i*3+1] = math.sqrt(self.masses[i])
            p3[i*3+2] = math.sqrt(self.masses[i])
        # end for i
        # Normalise
        p1 = p1 / math.sqrt(np.dot(p1, p1))
        p2 = p2 / math.sqrt(np.dot(p2, p2))
        p3 = p3 / math.sqrt(np.dot(p3, p3))
        # Form the projection operators
        P1 = unit - np.outer(p1, p1)
        P2 = unit - np.outer(p2, p2)
        P3 = unit - np.outer(p3, p3)
        # Now project out
        new_hessian = np.dot(np.dot(P1.T, hessian), P1)
        new_hessian = np.dot(np.dot(P2.T, new_hessian), P2)
        return np.dot(np.dot(P3.T, new_hessian), P3)

    def _read_till_phrase(self,phrase):
        """Read lines from the current file until a match with phrase is found.

        Once a match is found, return the matching line.

        Parameters
        ----------
        phrase : str
            The phrase to match in each line of the file.

        Returns
        -------
        str
            The line from the file that matches the phrase.

        """
        line = ""
        while not phrase.match(line):
            line = self.file_descriptor.readline()
        return line

    def _dynamical_matrix(self, hessian):
        """Process the dynamical matrix.

        Processes the dynamical matrix by performing several steps: symmetrizing the Hessian, which is a nxn matrix of mass-weighted force constants, projecting out translational modes, diagonalizing the Hessian, and finally storing the frequencies and normal modes.

        Parameters
        ----------
        hessian : nxn array of floats
            The second derivative matrix

        Returns
        -------
        None

        Notes
        -----
        - Hessian is a nxn matrix of the mass weighted force constants.
        - The hessian is symmetrised.
        - Translational modes are projected out.
        - The hessian is diagonalised.
        - Finally, the frequencies and normal modes are stored.

        """
        if self.debug:
            print("_dynamical_matrix")
            print("hessian", hessian[0:4][0])
        masses = np.array(self.masses)*amu
        if self.debug:
            print("masses", self.masses, masses)
        if not self.nomass_hessian_has_been_set:
            # symmetrise the hessian and store it for later use
            if self.hessian_symmetrisation == "symm":
                hessian = 0.5 * (hessian + hessian.T)
            else:
                # Crystal only uses the upper triangle to calculate the eigenvalues
                hessian = hessian - np.tril(hessian) + np.triu(hessian).T
            self.nomass_hessian_has_been_set = True
            self.nomass_hessian = self._remove_mass_weighting(hessian,masses)
        if self.debug:
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        hessian = self._modify_mass_weighting(self.nomass_hessian,masses)
        if self.debug:
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        if self.debug:
            print("mass weighted hessian", hessian[0:4][0])
        # Project out the translational modes if requested
        if self.eckart:
            hessian = self.project(hessian)
        if self.debug:
            print("projected hessian", hessian[0:4][0])
        self.mass_weighted_normal_modes, self.frequencies = calculate_normal_modes_and_frequencies(hessian)
        if self.debug:
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        return

    def reset_born_charges(self):
        """Reset the born charges to their original values if they are not currently being used.

        This method restores the `born_charges` attribute of the instance to the value stored in `original_born_charges`, but only if `original_born_charges_are_being_used` is `False`.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        This method changes the state of the reader by modifying its `born_charges` attribute to match `original_born_charges`, under the condition that `original_born_charges_are_being_used` is `False`.

        """        
        if not self.original_born_charges_are_being_used:
            self.born_charges = self.original_born_charges

    def neutralise_born_charges(self):
        """Neutralise Born charges within the object.

        Changes the state of `original_born_charges_are_being_used` 
        to False and saves the current `born_charges` as `original_born_charges` if
        `original_born_charges_are_being_used` is True, indicating that the 
        original Born charges are no longer being used directly. It then applies 
        the Born charge sum rule by calling the `_born_charge_sum_rule` method.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        This method is intended to be used within a context where Born charges 
        (representative of the polarization of ions in a solid under an electric field) 
        need to be neutralized or altered from their original state.

        """        
        if self.original_born_charges_are_being_used:
            self.original_born_charges = self.born_charges
            self.original_born_charges_are_being_used = False
        self._born_charge_sum_rule()
        return

    def _born_charge_sum_rule(self):
        """Apply a simple charge sum rule to all the elements of the born matrices.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        This function applies a straightforward summation rule to the elements of the born matrices,

        """
        total = np.zeros((3, 3))
        born_charges = np.array(self.original_born_charges)
        new_born_charges = np.zeros_like(self.born_charges)
        total = np.sum(born_charges) / self.nions
        if self.debug:
            print("born charge sum", total)
        new_born_charges = born_charges - total
        self.born_charges = new_born_charges.tolist()
        return

    def _modify_mass_weighting(self,hessian,new):
        """Modify the Hessian matrix based on new mass weighting.

        This function iterates over the elements of the Hessian matrix (`hessian`) and adjusts each element based on the square root of the product of elements from a new weighting (`new`). 

        Parameters
        ----------
        hessian : numpy.ndarray
            The original Hessian matrix whose elements are to be modified. It is assumed to be a square matrix with dimensions [3*nions, 3*nions], where nions is the number of ions (or atoms) in the system.
        new : numpy.ndarray or list
            An array or list of new weighting factors, one per ion. Its length should be equal to `nions`. These factors are applied in the modification of the Hessian matrix elements.

        Returns
        -------
        numpy.ndarray
            The modified Hessian matrix with the same dimensions as the input matrix.

        """        
        ipos = -1
        new_hessian = np.empty_like(hessian)
        for i in range(self.nions):
            for _ix in range(3):
                ipos += 1
                jpos = -1
                for j in range(self.nions):
                    for _jx in range(3):
                        jpos += 1
                        new_hessian[ipos,jpos] = hessian[ipos,jpos] / math.sqrt(  new[i]*new[j] )
                    # end for jx
                # end for j
            # end for ix
        # end for i
        return new_hessian

    def _remove_mass_weighting(self,hessian,old):
        """Remove mass-weighting from a Hessian matrix.

        Parameters
        ----------
        hessian : ndarray
            The Hessian matrix to be un-weighted, typically a square, 2D array where
            the dimension correlates with 3 times the number of ions (nions*3) since
            it includes the x, y, and z coordinates for each ion.
        old : ndarray or list
            An array or list of mass weights for each ion.

        Returns
        -------
        new_hessian : ndarray
            The mass-weighting removed (un-weighted) Hessian matrix.

        Notes
        -----
        This function assumes that the Hessian matrix is provided in a form that
        has been mass-weighted. Each element of the `hessian` input is divided by
        the square root of the product of the mass weights of the corresponding ions,
        adjusting it to an un-weighted form.

        The size of `old` must match the number of ions (`nions`). Each entry in `old`
        is used to compute the square root of the product of mass weights for the 
        appropriate matrix element, thereby removing the mass weighting from the 
        original Hessian matrix.

        """        
        new_hessian = np.empty_like(hessian)
        ipos = -1
        for i in range(self.nions):
            for _ix in range(3):
                ipos += 1
                jpos = -1
                for j in range(self.nions):
                    for _jx in range(3):
                        jpos += 1
                        new_hessian[ipos,jpos] = hessian[ipos,jpos] * math.sqrt( old[i]*old[j] )
                    # end for jx
                # end for j
            # end for ix
        # end for i
        return new_hessian



#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""The SuperCell module."""

import numpy as np

from PDielec.Plotter import print_ints, print_reals


class SuperCell:
    """Hold super-cell information; the unit cell and the super-cell size.

    Parameters
    ----------
    aUnitCell : :class:`~PDielec.UnitCell.UnitCell`
        The unit cell associated with this instance.
    anImageSpecifier : list or tuple
        The list or tuples of 3 numbers used to generate images.

    Attributes
    ----------
    unitCell : :class:`~PDielec.UnitCell.UnitCell`
        The unit cell associated with this instance. (see :class:`~PDielec.UnitCell.UnitCell`)
    imageList : NoneType or list
        Initialized as None, can later hold a list of images of the cell
    imageSpecifier : list or tuple
        A list or tuple of 3 numbers which indicate the number of images of the cell to be created
    XYZ : NoneType or array_like
        Initialized as None, can later hold XYZ coordinates. (Cartesian coordinates)

    Notes
    -----
    This method automatically invokes the createImages method at the end of initialization.

    See Also
    --------
    :class:`~PDielec.UnitCell.UnitCell`

    """

    def __init__(self, aUnitCell,anImageSpecifier):
        """Initialize the supercell with a unit cell and an image specifier.

        Parameters
        ----------
        aUnitCell : :class:`~PDielec.UnitCell.UnitCell`
            The unit cell associated with this instance.
        anImageSpecifier : list or tuple
            The list or tuples of 3 numbers used to generate images.

        Attributes
        ----------
        unitCell : :class:`~PDielec.UnitCell.UnitCell`
            The unit cell associated with this instance.
        imageList : NoneType or list
            Initialized as None, can later hold a list of images of the cell
        imageSpecifier : list or tuple
            A list or tuple of 3 numbers which indicate the number of images of the cell to be created
        XYZ : NoneType or array_like
            Initialized as None, can later hold XYZ coordinates. (Cartesian coordinates)

        Notes
        -----
        This method automatically invokes the createImages method at the end of initialization.

        """
        self.unitCell       = aUnitCell         # the unit cell on which everything is based
        self.imageList      = None              # a list of the images of the unit cell
        self.imageSpecifier = anImageSpecifier  # the specification used to create the imageList
        self.XYZ            = None              # Cartesian coordinates of the super-cell
        self.bonds          = []                # A list of bonds in the super-cell
        self.createImages(anImageSpecifier)
        self.calculateXYZ()

    def print(self):
        """Print the contents of the SuperCell.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Example
        --------
        ```
        # Create a 3x2x1 super-cell
        import PDielec.SuperCell as SC
        import PDielec.UnitCell as UC
        uc=UC.UnitCell(1.0,2.0,3.0,90.0,90.0,90.0)
        imageSpecifier = [3,2,1]
        sc=SC.SuperCell(uc,imageSpecifier)
        sc.print()
        ```

        """
        print("SuperCell")
        print_ints("Image_specifier",self.imageSpecifier)
        self.unitCell.print()
        print("SuperCell Image list")
        for l in self.imageList:
            print_ints("",l)
        corners,edges,labels = self.getBoundingBox()
        print("SuperCell Corners")
        for corner in corners:
            print_reals("",corner,format="{:12.6f}")
        if self.XYZ is not None:
            print()
            print("SuperCell XYZ (Coordinates)")
            for xyz in self.XYZ:
                print_reals("",xyz,format="{:12.6f}")
        if len(self.bonds) > 0:
            print()
            print("SuperCell Bond list")
            for bond in self.bonds:
                print_ints("",bond)

    def createImages(self,anImageSpecifier):
        """Create a list of images extending by ia, jb, kc units in the a, b, c directions. The image specifier is a list of the form [ia, jb, kc].

        Parameters
        ----------
        anImageSpecifier : list (ia, jb, kc)
            ia an integer specifying extension in the a direction.
            jb an integer specifying extension in the b direction.
            kc an integer specifying extension in the c direction.

        Returns
        -------
        list
            A list of created images.

        """
        ia,jb,kc = anImageSpecifier
        if ia < 1 or jb < 1 or kc < 1 :
            print("Error in createSuperCell: parameters must be 1 or more")
            return
        self.imageSpecifier = anImageSpecifier
        self.imageList = []
        for i in range(ia):
            for j in range(jb):
                for k in range(kc):
                    self.imageList.append( (i,j,k) )

    def getElementNames(self):
        """Calculate and return the element names of the supercell.

        Parameters
        ----------
        None

        Returns
        -------
        list
            The element names of the supercell.

        """
        names = []
        cell_names = self.unitCell.element_names
        for  _i,_j,_k in self.imageList:
            names.extend( cell_names )
        #end for i,j,k
        return names

    def calculateNormalModes(self,modes):
        """Calculate and return the normal modes of the supercell.

        Normal modes are initially an np array of nmodes, 3*nions.
        The number of modes stays the same but the number of ions
        increases according to the number of images.

        Parameters
        ----------
        modes : a list of floats
            A list of the modes to be used in the calculation

        Returns
        -------
        np.ndarray
            Array representing the normal modes of the supercell. Initially, this is an array
            of shape (nmodes, 3*nions), where `nmodes` is the number of modes and `nions` is the
            number of ions per mode. The shape may vary as the number of ions increases with the
            number of images.

        """
        nmodes,nions,_ = np.shape(modes)
        nions3 = nions*3
        # Make a copy of the modes, transforming [nmodes, nions, 3] to [ nmodes, nions3 ]
        modes_copy = np.zeros( (nmodes, nions3) )
        for imode,mode in enumerate(modes):
            pos = 0
            modes_copy[imode] = mode.flatten()
        # Now make multiple copies of the normal modes for each image.
        mult = len(self.imageList)
        normal_modes = np.zeros( (nmodes, mult*nions3) )
        for mode_index,mode in enumerate(modes_copy):
            pos = 0
            for  _i,_j,_k in self.imageList:
                for m in mode:
                  normal_modes[mode_index,pos] = m
                  pos += 1
                #end for m in mode
            #end for i,j,k
        #end for i,mode 
        return normal_modes

    def calculateXYZ(self):
        """Calculate and return the cartesian coordinates of the supercell.

        The fractional cell coordinates of the unit cell are used to calculate the XYZ coordinates of the super-cell

        Parameters
        ----------
        None

        Returns
        -------
        output : array
            The Cartesian coordinates of the supercell.

        """
        self.XYZ = []
        coords = []
        for  i,j,k in self.imageList:
            for fx,fy,fz in self.unitCell.fractional_coordinates:
                fx += i
                fy += j
                fz += k
                coords.append( (fx,fy,fz) )
            # end for fx,fy,fz
        #end for i,j,k
        self.XYZ = self.unitCell.convert_abc_to_xyz(coords)
        return self.XYZ

    def calculateBonds(self):
        """Calculate the bonding in the supercell.

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of the bonds as pairs of atom indexes

        Notes
        -----
        The bonding in the unit-cell is assumed to be valid whole-molecule bonding
        the bonds are not recalculated, just extended into the super-cell

        """
        self.bonds = []
        natoms = self.unitCell.nions
        bonds = self.unitCell.bonds
        for cell in range(len(self.imageList)):
            for na,nb in bonds:
              na += cell*natoms
              nb += cell*natoms
              self.bonds.append( (na,nb) )
            # end for na,nb
        #end for i,j,k
        return self.bonds


    def getBoundingBox(self,originABC=None):
        """Calculate the bounding box of an object in Cartesian coordinates based on its image specifications and an origin.

        Return a box with 8 corners and 12 edges which represent the supercel in cartesian space
        - Corners in abc space
        - shift the origin to originABC

        Parameters
        ----------
        originABC : array-like, optional
            The origin coordinates (A, B, C) to which the bounding box will be translated.
            defaults to 0,0,0

        Returns
        -------
        tuple
            A tuple containing three elements:
    
            - corners_xyz : list
              A list of corner points in XYZ coordinates specifying the bounding box, shifted to the provided origin.
    
            - edges : list
              A list of tuples, each representing an edge of the bounding box defined by a pair of XYZ coordinates.

            - labels : list of strings
              A list of strings labelling the bounding box a,b,c,o


        Notes
        -----
        The function calculates the center of the bounding box, re-centers the corners to the provided origin, and then builds edges between these corners in XYZ space.

        """
        if originABC is None:
            originABC = [0.0, 0.0, 0.0]
        i,j,k = self.imageSpecifier
        i = float(i)
        j = float(j)
        k = float(k)
        corners = []
        corners.append( np.array( [0.0,0.0,0.0] ) )
        corners.append( np.array( [  i,0.0,0.0] ) )
        corners.append( np.array( [  i,  j,0.0] ) )
        corners.append( np.array( [0.0,  j,0.0] ) )
        corners.append( np.array( [0.0,0.0,  k] ) )
        corners.append( np.array( [  i,0.0,  k] ) )
        corners.append( np.array( [  i,  j,  k] ) )
        corners.append( np.array( [0.0,  j,  k] ) )
        labels = [ "o", "a", "", "b", "c", "", "", "" ]
        centre = np.array ( [0.0, 0.0, 0.0] )
        for corner in corners:
            centre += corner
        centre = centre / 8.0
        corners = [ corner - centre + originABC for corner  in corners ]
        # Now calculate the corners in cartesian space
        corners_xyz = self.unitCell.convert_abc_to_xyz(corners)
        # Now calculate the edges each edge is a tuple with a beginning coordinate and an end coordinate
        edges = []
        edges.append( (corners_xyz[1] , corners_xyz[0]) )
        edges.append( (corners_xyz[2] , corners_xyz[1]) )
        edges.append( (corners_xyz[3] , corners_xyz[2]) )
        edges.append( (corners_xyz[3] , corners_xyz[0]) )
        edges.append( (corners_xyz[4] , corners_xyz[0]) )
        edges.append( (corners_xyz[5] , corners_xyz[1]) )
        edges.append( (corners_xyz[6] , corners_xyz[2]) )
        edges.append( (corners_xyz[7] , corners_xyz[3]) )
        edges.append( (corners_xyz[5] , corners_xyz[4]) )
        edges.append( (corners_xyz[6] , corners_xyz[5]) )
        edges.append( (corners_xyz[7] , corners_xyz[6]) )
        edges.append( (corners_xyz[7] , corners_xyz[4]) )
        return corners_xyz,edges,labels

    def calculateCentreOfBox(self):
        """Calculate the centre of the super-cell in cartesian coordinates.

        This is the centre of the box, not the centre of mass

        Parameters
        ----------
        None

        Returns
        -------
        vector
            The centre of the images in the box

        """
        i,j,k = self.imageSpecifier
        i = 0.5*float(i)
        j = 0.5*float(j)
        k = 0.5*float(k)
        ijk = [ i,j,k ]
        return self.unitCell.convert_abc_to_xyz(ijk)


    def calculateCentreOfMass(self,output="xyz"):
        """Calculate the center of mass for a molecular structure.

        Parameters
        ----------
        output : str, optional
            The output. Options are 'xyz' for mass and Cartesian coordinates,
            'mass' for the total mass only, 'abc' for mass and fractional coordinates, or any other string
            to return a tuple with all three values (mass, Cartesian coordinates, fractional coordinates).
            The default value is 'xyz'.

        Returns
        -------
        numpy.ndarray or float or tuple
            Depending on the value of `units`:
            If 'xyz', returns a tuple with mass and a numpy.ndarray of Cartesian coordinates of the center of mass.
            If 'mass', returns the total mass as a float.
            If 'abc', returns tuple with mass and a numpy.ndarray of fractional coordinates of the center of mass.
            For any other value of `output`, returns a tuple containing the total mass (float), Cartesian
            coordinates (numpy.ndarray), and fractional coordinates (numpy.ndarray) of the center of mass.

        """
        # Calculate the centre of mass 
        # The centre of mass can be returned as 'xyz' space or 'abc' space
        # if output='all' a tuple of (mass,cm_xyz,cm_abc) is returned
        # if output='mass' a mass is returned
        mass = 0.0
        cm_fractional = np.zeros(3)
        unit_mass, unit_cm_xyz, unit_cm_fractional = self.unitCell.calculateCentreOfMass(output=all)
        for  i,j,k in self.imageList:
            cm_fractional[0] += i + unit_cm_fractional [0]
            cm_fractional[1] += j + unit_cm_fractional [1]
            cm_fractional[2] += k + unit_cm_fractional [2]
            mass += unit_mass
        #end for i,j,k
        cm_fractional = cm_fractional/len(self.imageList)
        cm_xyz = self.unitCell.convert_abc_to_xyz(cm_fractional)
        return {
            "xyz"  : cm_xyz,
            "mass" : mass,
            "abc"  : cm_fractional,
            "all"  : (mass, cm_xyz, cm_fractional ),
        }.get(output, (mass, cm_xyz, cm_fractional))



#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""CastepOutputReader module."""

import os
import re

import numpy as np

from PDielec.GenericOutputReader import GenericOutputReader
from PDielec.UnitCell import UnitCell


class CastepOutputReader(GenericOutputReader):
    """Read the contents of a Castep output file.

    Inherits from :class:`~PDielec.GenericOutputReader.GenericOutputReader`

    Parameters
    ----------
    filenames : list
        A list of filenames (strings) intended for initialization. The first filename in the list is used to determine the seedname and the types of files to be created (.castep and .phonon).

    Notes
    -----
    - This method assumes that the filenames provided in the list relate to CASTEP calculation outputs. Specifically, it looks for filenames ending in '.castep' or '.phonon' to set up the necessary output files for the instance.
    - The method does not return a value but initializes the instance with the necessary attributes for further processing and analysis of CASTEP output files.

    """

    def __init__(self, filenames):
        """Initialize a new instance of the class CastepOutputReader.

        This constructor initializes the class with data from the provided filenames. It sets up the necessary properties for further operations, including identifying and setting up key output files (.castep and .phonon files) from the given filenames. Additional properties related to electronic calculations, phonon calculations, and ionic types are also initialized but not explicitly defined.

        Parameters
        ----------
        filenames : list
            A list of filenames (strings) intended for initialization. The first filename in the list is used to determine the seedname and the types of files to be created (.castep and .phonon).

        Notes
        -----
        - This method assumes that the filenames provided in the list relate to CASTEP calculation outputs. Specifically, it looks for filenames ending in '.castep' or '.phonon' to set up the necessary output files for the instance.
        - The method does not return a value but initializes the instance with the necessary attributes for further processing and analysis of CASTEP output files.

        Raises
        ------
        ValueError
            If the first filename does not contain '.castep' or '.phonon', implying it may not be suitable for initializing this class instance as intended. (Note: This exception raising is implied and should ideally be added to the code to handle cases where filenames do not meet expected criteria.)

        """        
        GenericOutputReader.__init__(self, filenames)
        if filenames[0].find(".castep") or filenames[0].find(".phonon"):
            seedname, ext = os.path.splitext(filenames[0])
        self._castepfile             = seedname+".castep"
        self._phononfile             = seedname+".phonon"
        self.names                   = [self._castepfile, self._phononfile]
        self._outputfiles            = [self._castepfile, self._phononfile]
        self.type                    = "Castep output"
        # Specific Castep Reader Variables
        self._pspots                  = {}
        self._ediff                   = 0.0
        self._epsilon                 = None
        self._nbranches              = 0
        self._pulay                  = None
        self._ion_type_index         = {}
        self._ion_index_type         = {}
        self._intensities             = None
        self.de_ion                  = []
        self.fmax                    = []
        self.dr_max                  = []
        self.smax                    = []
        return

    def _read_output_files(self):
        """Define the strings needed for searching the files process the output files.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.manage = {}   # Empty the dictionary matching phrases
        self.manage["spin"]          = (re.compile(" *net spin   of"), self._read_spin)
        self.manage["nelect"]        = (re.compile(" *number of  electrons"), self._read_nelect)
        self.manage["cellcontents"]  = (re.compile(" *Unit Cell"), self._read_cellcontents)
        self.manage["pspots"]        = (re.compile(" *Files used for pseudopotentials:"), self._read_pspot)
        self.manage["masses"]        = (re.compile(" *Mass of species in AMU"), self._read_masses)
        self.manage["kpoints"]       = (re.compile(" *Number of kpoints used"), self._read_kpoints)
        self.manage["kpoint_grid"]   = (re.compile(" *MP grid size for SCF"), self._read_kpoint_grid)
        self.manage["energies"]   = (re.compile(" *Final energy, E"), self._read_energies)
        self.manage["energies2"]   = (re.compile("Final energy ="), self._read_energies2)
        self.manage["energies3"]        = (re.compile("Dispersion corrected final energy"), self._read_energies3)
        self.manage["energy_cutoff"] = (re.compile(" *plane wave basis set cut"), self._read_energy_cutoff)
        self.manage["convergence"] = (re.compile(".*finished iteration"), self._read_convergence)
        self.manage["nbands"]        = (re.compile(" *number of bands"), self._read_nbands)
        self.manage["pressure"]      = (re.compile(" *\\* *Pressure: "), self._read_external_pressure)
        self.manage["opticalDielectric"]  = (re.compile(" *Optical Permittivity"), self._read_dielectric)
        self.manage["bornCharges"]    = (re.compile(" *Born Effective Charges"), self._read_born_charges)
        #  For the .phonon file
        self.manage["frequency"]      = (re.compile("     q-pt=    1    0.000000  0.000000  0.000000      1.0000000000 *$"), self._read_frequencies)
        self.manage["nbranches"]      = (re.compile(" Number of branches"), self._read_nbranches)
        for f in self._outputfiles:
            self._read_output_file(f)
        return

    def _read_nbranches(self, line):
        """Parse the number of branches from a line of text from the phonon file and set it to an instance variable.

        Parameters
        ----------
        line : str
            The line of text containing the number of branches information. 
            It is expected that the number of branches can be found at the fourth position 
            (index 3) when splitting the line by whitespace.

        Returns
        -------
        None

        Notes
        -----
        This function does not return anything as it sets the parsed number directly to an instance variable named `_nbranches`.

        """        
        self._nbranches = int(line.split()[3])

    def _read_frequencies(self, line):
        """Read and process frequencies, intensities, and normal modes from  phonon file.

        This method assumes a specific format of the input file, where frequencies, intensities, and normal mode coordinates are listed in a sequential order. Frequencies and intensities are read first, followed by normal mode vectors for each mode.

        Parameters
        ----------
        line : str
            The current line in the file from which to start reading the frequency data.

        Returns
        -------
        None

        Attributes Modified
        -------------------
        mass_weighted_normal_modes : list
            A list of mass-weighted normal modes. Each normal mode is represented as a list of 3D vectors (lists) for each ion in the structure.
        frequencies : list
            A list of frequency values read from the file.
        _intensities : list
            A list of intensity values corresponding to each frequency.

        Notes
        -----
        - This method directly modifies the instance attributes `mass_weighted_normal_modes`, `frequencies`, and `_intensities`.
        - It is assumed that `self.file_descriptor` is an open file object from which the data is read.
        - The method relies on the internal variable `self._nbranches` to determine the number of modes to read, and `self.nions` for the number of ions per mode.

        """        
        self.mass_weighted_normal_modes = []
        # maybe should use _nbranches
        frequencies = []
        intensities = []
        normal_modes = []
        for _imode in range(self._nbranches):
            line = self.file_descriptor.readline()
            frequencies.append(float(line.split()[1]))
            intensities.append(float(line.split()[2]))

        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        for _imode in range(self._nbranches):
            a = []
            for _ion in range(self.nions):
                line = self.file_descriptor.readline()
                a.append([float(line.split()[2]), float(line.split()[4]), float(line.split()[6])])
            # end for ion
            normal_modes.append(a)
        # end for imode
        self.frequencies = []
        self._intensities = []
        self.mass_weighted_normal_modes = []
        # now reads all frequencies imaginary or not
        # imaginary frequencies are indicated by real negative values
        for i, freq in enumerate(frequencies):
            self.frequencies.append(freq)
            self._intensities.append(intensities[i])
            self.mass_weighted_normal_modes.append(normal_modes[i])
            # end of if freq
        # end of for freq
        return

    def _read_kpoint_grid(self, line):
        """Parse and set the k-point grid dimensions from the given line.

        This method updates the `kpoint_grid` attribute of the object with the k-point grid dimensions obtained from parsing the specified line. The k-point grid dimensions are expected to be located at the 8th, 9th, and 10th positions (1-based indexing) in the line, separated by spaces.

        Parameters
        ----------
        line : str
            The line from which the k-point grid dimensions are to be extracted.

        Returns
        -------
        None

        """        
        self.kpoint_grid = [ int(line.split()[7]), int(line.split()[8]), int(line.split()[9]) ]
        return

    def _read_kpoints(self, line):
        """Extract number of k-points from a given line of text and update the instance variable.

        Parameters
        ----------
        line : str
            The line of text from which the number of k-points is extracted.

        Returns
        -------
        None

        Notes
        -----
        This function assumes the number of k-points is always located at the 6th position (index 5) when the line is split by whitespace. It directly updates the `kpoints` attribute of the class instance.

        """        
        self.kpoints = int(line.split()[5])
        return

    def _read_nbands(self, line):
        """Extract the number of bands from a given line and set it to the 'nbands' attribute of the class.

        Parameters
        ----------
        line : str
            The line from which the number of bands is to be extracted. It is expected that the fifth element (index 4) when splitting the line by spaces is an integer representing the number of bands.

        Returns
        -------
        None

        """        
        self.nbands = int(line.split()[4])
        return

    def _read_pseudoatom(self, line):
        """Read and process a pseudoatom line from a data file.

        Parameters
        ----------
        line : str
            A string representing a line from a data file that contains information about a pseudoatom.

        Returns
        -------
        None

        Notes
        -----
        This method processes a line from a data file, extracting the species from the sixth (index 5) position in the whitespace-separated list. The species name is capitalized and stored, and indices related to the species and ion types are updated accordingly.

        """        
        species = line.split()[5].capitalize()
        # These are two private dictionary to map the species name to a type index
        self.species.append(species)
        self._ion_type_index[species] = self.nspecies
        self._ion_index_type[self.nspecies] = species
        self.nspecies += 1
        return

    def _read_cellcontents(self, line):
        """Read the cell contents from an input file and updates the class attributes accordingly.

        This method reads the real lattice vectors, current cell volume, total number of ions, 
        species, and fractional coordinates from the input file, and updates the relevant class 
        attributes including `unit_cells`, `volume`, `volumes`, `nions`, `nspecies`, 
        `atom_type_list`, `species`, `ions_per_type`, along with setting the fractional 
        coordinates and element names for the latest unit cell.

        Parameters
        ----------
        line : str
            The initial line from which to start reading. This parameter is not used in the 
            current implementation and could be omitted or replaced depending on the method's 
            usage context.

        Returns
        -------
        None

        """        
        line = self._read_till_phrase(re.compile(" *Real Lattice"))
        line = self.file_descriptor.readline()
        avector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        bvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        cvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        self.unit_cells.append(UnitCell(avector, bvector, cvector,units="Angstrom"))
        self.ncells = len(self.unit_cells)
        line = self._read_till_phrase(re.compile(" *Lattice*"))
        line = self._read_till_phrase(re.compile(" *Current cell volume"))
        self.volume = float(line.split()[4])
        self.volumes.append(float(line.split()[4]))
        if self.nions == 0:
            line = self._read_till_phrase(re.compile(" *Total number of ions in cell"))
            if len(self.unit_cells) == 1:
                self.nions = int(line.split()[7])
                line = self.file_descriptor.readline()
                self.nspecies = int(line.split()[7])
                line = self.file_descriptor.readline()
                line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
        # end if self.nions
        else:
            # Not all Unit Cell output is the same in Castep 17
            line = self._read_till_phrase(re.compile(" *xxxxxxxxxxxxxxxxxxxxxxxxxxxx"))
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
        # end of else
        fractional_coordinates = []
        # ions_per_type is a dictionary here and is local
        ions_per_type = {}
        self.atom_type_list = []
        self.species = []
        species_list = []
        for _i in range(self.nions):
            line = self.file_descriptor.readline()
            atom_type = line.split()[1].capitalize()
            species_list.append(atom_type)
            if atom_type not in ions_per_type:
                self.species.append(atom_type)
                ions_per_type[atom_type] = 0
            self.atom_type_list.append(self.species.index(atom_type))
            ions_per_type[atom_type] += 1
            atom_frac = [float(f) for f in line.split()[3:6]]
            fractional_coordinates.append(atom_frac)
        # At some point we should store the fractional coordinates
        for species in self.species:
            n = ions_per_type[species]
            # self.ions_per_type is a list
            self.ions_per_type.append(n)
        self.unit_cells[-1].set_fractional_coordinates(fractional_coordinates)
        self.unit_cells[-1].set_element_names(species_list)
        return

    def _read_masses(self, line):
        """Read and parse masses from a file, categorizing them by species.

        This method reads lines from an opened file associated with the instance, extracting the species name, its mass, and the number of ions per species. It updates instance attributes to store this information, organizing masses both by type and in total.

        Parameters
        ----------
        line : str
            The initial line from which to start reading in the file. This parameter is part of the method signature for demonstration but is immediately overwritten by the next line read from the file, hence not directly used.

        Returns
        -------
        None

        """        
        self.masses = []
        self.masses_per_type = []
        self.species = []
        self.nspecies = 0
        line = self.file_descriptor.readline()
        while len(line.split()) != 0:
            species = line.split()[0].capitalize()
            mass = float(line.split()[1])
            nions = self.ions_per_type[self.nspecies]
            self.species.append(species)
            # These are two private dictionary to map the species name to a type index
            # self._ion_type_index[species] = self.nspecies
            self._ion_index_type[self.nspecies] = species
            self.masses_per_type.append(mass)
            self.nspecies += 1
            for _j in range(nions):
                self.masses.append(mass)
            line = self.file_descriptor.readline()
        # end while loop
        return

    def _read_born_charges(self, line):
        """Read the born charges from the castep file.

        This function reads the born charges from a provided castep file and arranges the output tensor. Initially, each column of the output refers to a given field direction, and each row refers to the atomic displacement, organizing the numbers in the format `[[a1x a2x a3x], [a1y a2y a3y], [a1z a2z a3z]]`. The function then rearranges these numbers into the desired format: `[[a1x a1y a1z], [a2x a2y a2z], [a3x a3y a3z]]`, where 1, 2, 3 are the field directions, and x, y, z represent the atomic displacements.

        Parameters
        ----------
        line : str
            The initial line from which to start reading in the file. This parameter is part of the method signature for demonstration but is immediately overwritten by the next line read from the file, hence not directly used.

        Returns
        -------
        ndarray
            An array where each row represents the born charges for a particular atomic displacement arranged in the format `[[a1x a1y a1z], [a2x a2y a2z], [a3x a3y a3z]]`.

        """
        line = self.file_descriptor.readline()
        self.born_charges = []
        for _i in range(self.nions):
            b = []
            line = self.file_descriptor.readline()
            b.append([float(f) for f in line.split()[2:5]])
            line = self.file_descriptor.readline()
            b.append([float(f) for f in line.split()[0:3]])
            line = self.file_descriptor.readline()
            b.append([float(f) for f in line.split()[0:3]])
            B = np.array(b)
            C = B.T
            self.born_charges.append(C.tolist())
        return

    def _read_dielectric(self, line):
        """Read and parse dielectric data from a file.

        This method reads dielectric data from the currently opened file associated with the object. It updates the object's properties for zero-frequency optical and static dielectric constants.

        Parameters
        ----------
        line : str
            The initial line from which to start reading. Note: this parameter is actually unused in the function's current implementation, and reading starts from the next line of the file descriptor's current position.

        Returns
        -------
        None

        Notes
        -----
        The function assumes that the file's current position is at the correct start point for reading dielectric data. It reads three consecutive lines from the file, each expected to contain six floating point numbers. The first three numbers of each line are appended to 'zerof_optical_dielectric', and the latter three numbers to 'zerof_static_dielectric'. 

        This method modifies the state of the object by updating the 'zerof_optical_dielectric' and 'zerof_static_dielectric' lists with new data read from the file. 

        """        
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        # this is epsilon infinity
        self.zerof_optical_dielectric = []
        # this is the zero frequency static dielectric constant
        self.zerof_static_dielectric = []
        self.zerof_optical_dielectric.append([float(f) for f in line.split()[0:3]])
        self.zerof_static_dielectric.append([float(f) for f in line.split()[3:6]])
        line = self.file_descriptor.readline()
        self.zerof_optical_dielectric.append([float(f) for f in line.split()[0:3]])
        self.zerof_static_dielectric.append([float(f) for f in line.split()[3:6]])
        line = self.file_descriptor.readline()
        self.zerof_optical_dielectric.append([float(f) for f in line.split()[0:3]])
        self.zerof_static_dielectric.append([float(f) for f in line.split()[3:6]])
        return

    def _read_skip4(self, line):
        """Read and skip the next four lines from the current file position.

        This method reads and discards the next four lines from the file
        associated with this object's file_descriptor attribute.

        Parameters
        ----------
        line : str
            This parameter is not used in the function body and serves no purpose
            in the context of this method's operation.

        Returns
        -------
        None

        """        
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        return

    def _read_external_pressure(self, line):
        """Parse an external pressure value from a string and set it.

        Parameters
        ----------
        line : str
            A string containing the pressure value to parse and set. The expected format is unspecified in this context, but the pressure value should be at the third position (index 2) when the string is split by whitespace.

        Returns
        -------
        None

        Notes
        -----
        This function updates the `pressure` attribute of the class instance based on the value parsed from the input string. The specific use case and format of the input string are not detailed, assuming some domain-specific knowledge is required.

        """        
        self.pressure = float(line.split()[2])
        return

    def _read_pspot(self, line):
        """Read potential spots from a file and store them.

        This method reads a specific number of lines determined by the `nspecies` attribute from
        the file associated with `file_descriptor`, processes them to extract potential spot
        identifiers and their corresponding values, and stores them in the `_pspots` attribute.

        Parameters
        ----------
        line : str
            The line from which the reading starts. This initial line is often a header or
            an indicator that the following lines contain the potential spots data.

        Returns
        -------
        None

        """        
        for _i in range(self.nspecies):
            line = self.file_descriptor.readline()
            self._pspots[line.split()[0]] = line.split()[1]
        return

    def _read_spin(self, line):
        """Read and set the spin value from a given line of text.

        The spin is extracted from the sixth element (index 5) of the splitted line.

        Parameters
        ----------
        line : str
            The input line from which the spin is extracted. Expected to have at least six space-separated elements, with the sixth element being a string representation of the spin value.

        Returns
        -------
        None

        """        
        self.spin = int(float(line.split()[5]))
        return

    def _read_energy_cutoff(self, line):
        """Extract the energy cutoff value from a given line and store it in the object.

        This method reads a line of text, expects to find a numeric value at the 7th position (0-indexed) in a space-separated list, converts this value to a float, and stores it as the energy cutoff property of the object.

        Parameters
        ----------
        line : str
            A string containing the energy cutoff value at the 7th position in a space-separated list.

        Returns
        -------
        None

        """        
        self.energy_cutoff = float(line.split()[6])
        return

    def _read_ediff(self, line):
        """Read and set the energy difference from a given line of text.

        This is a private method intended to parse and store the energy difference
        from a standardized formatted line of text.

        Parameters
        ----------
        line : str
            A line of text, expected to contain numerical values from which
            the energy difference (ediff) is parsed. The value for ediff is
            expected to be the third element when splitting the line by whitespace.

        Returns
        -------
        None

        """        
        self._ediff = float(line.split()[2])
        return

    def _read_nelect(self, line):
        """Extract the number of electrons from a line and assign it.

        Extracts the number of electrons from a string (line) by parsing it and assigns the value to the instance variable `electrons`.

        Parameters
        ----------
        line : str
            A string (typically a line from a file) that contains the number of electrons among other text, expected to be found at the fifth position (index 4) when the string is split.

        Returns
        -------
        None

        """        
        self.electrons = int(float(line.split()[4]))
        return

    def _read_energies(self, line):
        """Read and store energy values from a given line.

        Extracts energy-related data from a specified line and updates relevant attributes of the object. Assumes that the sixth (index 5) element in the second line contains the final free energy, the fifth (index 4) element in the given line contains the DFT energy, which is appended to a list, and also set as the final energy without entropy. Additionally, updates the number of geometry steps based on the length of the energiesDFT list.

        Parameters
        ----------
        line : str
            The line from which energy values are extracted.

        Returns
        -------
        None

        Notes
        -----
        This method expects that 'line' follows a specific format, where the fifth (index 4) element is the final energy without entropy and, on the subsequent line, the sixth (index 5) element is the final free energy. It further assumes 'self.file_descriptor' to be an open file object positioned at the line immediately following the line containing energy information.

        The method updates four attributes of the object:
        - Appends the energy value to 'self.energiesDFT'.
        - Sets 'self.final_energy_without_entropy' to the extracted energy value.
        - Reads the next line and sets 'self.final_free_energy' to the extracted value.
        - Updates 'self.geomsteps' to represent the total number of energy steps recorded.

        """        
        self.energiesDFT.append(float(line.split()[4]))
        self.final_energy_without_entropy = float(line.split()[4])
        line = self.file_descriptor.readline()
        self.final_free_energy = float(line.split()[5])
        self.geomsteps = int(len(self.energiesDFT))
        return

    def _read_energies2(self, line):
        """Parse a line of text to read energies and update properties.

        This method reads energies from a given line of text and updates various energy
        related properties of the object.

        Parameters
        ----------
        line : str
            A line of text that contains energy values.

        Notes
        -----
        This method specifically updates the `energiesDFT`, `final_energy_without_entropy`,
        `final_free_energy`, and `geomsteps` attributes of the object based on the energy
        value parsed from the given line. The energy value is expected to be located at
        the fourth position (index 3) when the line is split by spaces.

        Returns
        -------
        None

        """        
        self.energiesDFT.append(float(line.split()[3]))
        self.final_energy_without_entropy = float(line.split()[3])
        self.final_free_energy = float(line.split()[3])
        self.geomsteps = int(len(self.energiesDFT))
        return

        
    def _read_energies3(self, line):
        """Read energy values from a line of text and append to class attributes.

        This method parses a line of text, extracts the energy value from the sixth
        position (index 5), converts it into a float, and appends it to the class's
        energiesDFT_disp attribute. It also updates the class attributes for the final
        energy without entropy, the final free energy, and the total number of geometry
        steps (geomsteps) based on the current energy readings.

        Parameters
        ----------
        line : str
            A string containing energy information, expected to have at least six
            words with the energy value at the sixth position.

        Returns
        -------
        None

        Notes
        -----
        This method directly modifies the attributes of the class instance it belongs
        to. It does not return any value.

        """        
        self.energiesDFT_disp.append(float(line.split()[5]))
        self.final_energy_without_entropy = float(line.split()[5])
        self.final_free_energy = float(line.split()[5])
        self.geomsteps = int(len(self.energiesDFT_disp))
        return
        
    def _read_convergence(self, line):
        """Read convergence data from a file and update internal lists.

        This internal method reads specific lines from a file that is already opened and updates internal lists for de-ionization energy, maximum force (fmax), maximum displacement (dr_max), and maximum stress (smax) based on the values found in those lines.

        Parameters
        ----------
        line : str
            Not used but overwritten

        Returns
        -------
        None

        Notes
        -----
        This function assumes that the file being read follows a specific format where the required values are found at known line intervals and in a consistent column position (index 3 after splitting the line). 

        The function does not return any value, but it updates the internal state of the object it is a part of.

        """        
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.de_ion.append(float(line.split()[3]))
        line = self.file_descriptor.readline()
        self.fmax.append(float(line.split()[3]))
        line = self.file_descriptor.readline()
        self.dr_max.append(float(line.split()[3]))
        line = self.file_descriptor.readline()
        self.smax.append(float(line.split()[3]))
        line = self.file_descriptor.readline()
        return    
        
            




#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a directory containing Phonopy input and output files."""

import sys

import numpy as np

from PDielec.Constants import thz2cm1
from PDielec.GenericOutputReader import GenericOutputReader
from PDielec.UnitCell import UnitCell


class PhonopyOutputReader(GenericOutputReader):
    """Read the contents of a directory containing Phonopy input and output files.

    Inherits from :class:`~PDielec.GenericOutputReader.GenericOutputReader`

    Parameters
    ----------
    names : list
        A list of file names to be used.

    """

    def __init__(self, names):
        """Initialise a Phonopy output reader.

        Parameters
        ----------
        names : list
            A list of file names to be used.

        This routine reads files with fixed names:
                  phonopy.yaml    contains details of the cells used
                  qpoints.yaml    contains the dynamical matrix
                  BORN_PDIELEC    generated from the BORN file using phonopy-pdielec-born

        """        
        GenericOutputReader.__init__(self, names)
        self.type                    = "Phonopy output"
        self.units                   = { "length"          : "angstrom",
                                         "mass"            : "AMU",
                                         "force"           : None,
                                         "force_constants" : None, }
        return

    def _read_output_files(self):
        """Read the Phonopy files in the directory.

        reads the dynamical matrix from phonopy
        reads the Born charges
        """
        # Calculate dynamical matrix
        qpoints_filename = self._outputfiles[0]
        phonopy_filename = self._outputfiles[1]
        born_filename    = self._outputfiles[2]
        self.read_dynamical_matrix(phonopy_filename,qpoints_filename)
        self.read_born_file(self.nions,born_filename)
        return

    def read_dynamical_matrix(self,phonopy_filename, qpoints_filename):
        """Read and process the dynamical matrix from output files.

        This method reads the dynamical matrix and other relevant data from the specified
        output files, calculates the Hessian matrix, converts its units, and computes the
        eigenvalues and eigenvectors to determine frequencies and mass-weighted normal modes.

        Parameters
        ----------
        phonopy_filename : str
            The phonopy.yaml file name
        qpoints_filename : str
            The qpoints.yaml file name

        Returns
        -------
        None
            This method directly modifies the instance attributes `_old_masses`, `frequencies`, 
            and `mass_weighted_normal_modes` based on the data read and calculations performed.

        Raises
        ------
        FileNotFoundError
            If any of the specified output files cannot be found or opened.
        YAMLError
            If there is an error parsing the YAML files.
        MemoryError
            If there is insufficient memory to compute the eigendecomposition.

        Notes
        -----
        - It first attempts to use the CLoader for parsing YAML files for performance reasons and falls
          back to the standard Loader if CLoader is not available, issuing a warning in such cases.
        - The dynamical matrix is processed to compute real-valued Hessian by considering only the real
          parts. It is then converted to the desired units before computing the eigenvalues and eigenvectors.
        - The method computes frequencies in THz by first converting the eigenvalues from the Hessian to
          the correct units, then calculates the square root of their absolute values, preserving the sign
          to distinguish between real and imaginary modes.
        - Mass-weighted normal modes are then computed for each ion based on the eigenvectors, and the 
          frequencies and normal modes are stored in corresponding instance attributes.

        """        
        #
        # Yaml imports of large files are really slow....
        # Attempt to use the PyYaml C parser, using yaml.CLoader
        #
        import yaml
        try:
            from yaml import CLoader as Loader
        except Exception:
            print("WARNING: Yaml CLoader is not avaiable, using fallback",file=sys.stderr)
            from yaml import Loader as Loader
        # the first name has to be the qpoints file
        with open(qpoints_filename) as fd:
            data_q = yaml.load(fd, Loader=Loader)
        # the second name has to be the phonopy file
        with open(phonopy_filename) as fd:
            data_p = yaml.load(fd, Loader=Loader)
        #
        # process phonopy.yaml
        #
        try:
            conversion_factor_to_THz = data_p["phonopy"]["frequency_unit_conversion_factor"]
        except Exception:
            conversion_factor_to_THz = 15.633302
        conversion_factor_to_cm1 = conversion_factor_to_THz * thz2cm1
        #
        # Get units
        #
        if "physical_unit" in data_p:
            if "atomic_mass" in data_p["physical_unit"]:
                self.units["atomic_mass"]     = data_p["physical_unit"]["atomic_mass"]
            if "length" in data_p["physical_unit"]:
                self.units["length"]          = data_p["physical_unit"]["length"]
            if "force" in data_p["physical_unit"]:
                self.units["force"]           = data_p["physical_unit"]["force"]
            if "force_constants" in data_p["physical_unit"]:
                self.units["force_constants"] = data_p["physical_unit"]["force_constants"]
        #
        # Determine axes and cells
        #
        if "primitive_matrix" in data_p:
            self.primitive_transformation = data_p["primitive_matrix"]
        else:
            self.primitive_transformation = None
        primitive_cell = self.read_cell(data_p["primitive_cell"]) if "primitive_cell" in data_p else None
        #
        # Use the cell that is consistent with the primitive_axes
        #
        cell = primitive_cell
        self._old_masses = cell.get_atomic_masses()
        self.nions = len(self._old_masses)
        self.species = cell.get_species()
        self.nspecies = len(self.species)
        elements = cell.get_element_names()
        # Calculate the number of ions / type
        # Determine the atom type list
        self.ions_per_type = [ 0 for _species in self.species ]
        for element in elements:
            species_index = self.species.index(element)
            self.ions_per_type[species_index] += 1
            self.atom_type_list.append(species_index)
        # Determine the mass of each type
        self.mass_per_type = []
        for species in self.species:
             element_index = elements.index(species)
             self.masses_per_type.append(self._old_masses[element_index])
        # Proceed
        self.unit_cells = [ cell ]
        self.ncells = 1
        self.volume = cell.getVolume("Angstrom")
        #
        # Process qpoints.yaml
        #
        dynmat = []
        dynmat_data = data_q["phonon"][0]["dynamical_matrix"]
        for row in dynmat_data:
            vals = np.reshape(row, (-1, 2))
            dynmat.append(vals[:, 0] + vals[:, 1] * 1j)
        dynmat = np.array(dynmat)
        # Make sure the hessian is real
        hessian = np.real(dynmat)
        # We need to convert to cm-1
        hessian = hessian * conversion_factor_to_cm1 * conversion_factor_to_cm1
        # Find its eigenvalues and eigen vectors
        eig_val, eig_vec = np.linalg.eigh(hessian)
        self.mass_weighted_normal_modes = []
        nmodes = 3*self.nions
        # Store the new frequencies, using the negative convention for imaginary modes
        frequencies_a = np.sqrt(np.abs(eig_val.real)) * np.sign(eig_val.real)
        self.frequencies = frequencies_a.tolist()
        # Store the mass weighted normal modes
        for i in range(nmodes):
            mode = []
            n = 0
            for _j in range(self.nions):
                modea = [eig_vec[n][i], eig_vec[n+1][i], eig_vec[n+2][i]]
                n = n + 3
                mode.append(modea)
            self.mass_weighted_normal_modes.append(mode)
        # end for i
        return

    def read_cell(self, yaml):
        """Read and process yaml string and extract a unit cell.

        Parameters
        ----------
        yaml : str
            A yaml string containing a phonopy cell

        Returns
        -------
        cell : unitCell

        """        
        lattice = yaml["lattice"]
        points = yaml["points"]
        symbols = [ atom["symbol"] for atom in points ]
        coordinates = [ atom["coordinates"] for atom in points ]
        masses = [ atom["mass"] for atom in points ]
        cell = UnitCell(lattice[0], lattice[1], lattice[2], units=self.units["length"])
        cell.set_fractional_coordinates(coordinates)
        cell.set_element_names(symbols)
        cell.set_atomic_masses(masses)
        return cell

    def read_born_file(self,natoms,filename):
        """Read the BORN_PDIELEC file in the current directory.

        The BORN_PDIELEC file was created by a phonopy helper routine such as phonopy-pdielec-born
        It contains the optical permittivity and the Born charges

        Parameters
        ----------
        natoms : int
            The number of atoms
        filename : str
            The filename (probably "BORN_PDIELEC")

        Modifies
        --------
        zerof_optical_dielectric : the zero frequency optical permittivity
        born_charges             : the born charges

        """        
        with open(filename) as fd:
            #
            # Skip a line
            #
            line = fd.readline()
            #
            # Read dielectric constant
            #
            line = fd.readline().split()
            if len(line) != 9:
                print("BORN file format of line 2 is incorrect")
                return
            self.zerof_optical_dielectric = np.reshape([float(x) for x in line], (3, 3))
            #
            # Read Born effective charge
            #
            self.born_charges = np.zeros((natoms, 3, 3), dtype="double")
            for i in range(natoms):
                line = fd.readline().split()
                if len(line) == 0:
                    print("Number of lines for Born effect charge is not enough.")
                    return
                if len(line) != 9:
                    print("BORN file format of line %d is incorrect" % (i + 3))
                    return
                self.born_charges[i] = np.reshape([float(x) for x in line], (3, 3))
        return
